"""
Data handler and main window creator for 2D data inspection
"""
import os
import time
from copy import deepcopy

import numpy as np
from pyqtgraph.Qt import QtWidgets
from PyQt5.QtWidgets import QMessageBox

import piva.arpys_wp as wp
import piva.data_loader as dl
import piva.imageplot as ip
from piva.cmaps import cmaps
from piva.edc_fitter import EDCFitter
from piva.mdc_fitter import MDCFitter

app_style = """
QMainWindow{background-color: rgb(64,64,64);}
QTabWidget{background-color: rgb(64,64,64);}
"""
DEFAULT_CMAP = 'viridis'
ORIENTLINES_LINECOLOR = (164, 37, 22, 255)
erg_ax = 2
scan_ax = 0

class DataHandler2D :
    """ Object that keeps track of a set of 2D data and allows
    manipulations on it. In a Model-View-Controller framework this could be
    seen as the Model, while :class:`MainWindow <data_slicer.pit.MainWindow>`
    would be the View part.
    """

    def __init__(self, main_window):
        self.main_window = main_window
        self.binning = False

        # Initialize instance variables
        # np.array that contains the 3D data
        self.data = None
        self.axes = np.array([[0, 1], [0, 1], [0, 1]])
        # Indices of *data* that are displayed in the main plot
        self.displayed_axes = (0, 1)
        # Index along the z axis at which to produce a slice
        # self.z = TracedVariable(0, name='z')
        # # Number of slices to integrate along z
        # integrate_z = TracedVariable(value=0, name='integrate_z')
        # How often we have rolled the axes from the original setup
        # self._roll_state = 0

        # moved to get rid of warnings
        # self.zmin = 0
        # self.zmax = None
        # self.integrated = None
        # self.cut_y_data = None
        # self.cut_x_data = None

    def get_data(self):
        """ Convenience `getter` method. Allows writing ``self.get_data()``
        instead of ``self.data.get_value()``.
        """
        if self.main_window.util_panel.image_normalize_edcs.isChecked():
            return self.norm_data
        else:
            return self.data.get_value()[0, :, :]

    def set_data(self, data):
        """ Convenience `setter` method. Allows writing ``self.set_data(d)``
        instead of ``self.data.set_value(d)``.
        """
        self.data.set_value(data)

    def prepare_data(self, data, axes=3 * [None]):
        """ Load the specified data and prepare the corresponding z range.
        Then display the newly loaded data.

        **Parameters**

        ====  ==================================================================
        data  3d array; the data to display
        axes  len(3) list or array of 1d-arrays or None; the units along the
              x, y and z axes respectively. If any of those is *None*, pixels
              are used.
        ====  ==================================================================
        """

        self.data = ip.TracedVariable(data, name='data')
        self.axes = np.array(axes, dtype="object")
        self.norm_data = wp.normalize(self.data.get_value()[0, :, :].T).T

        self.prepare_axes()
        self.on_z_dim_change()

        self.main_window.update_main_plot()
        self.main_window.set_axes()

    def prepare_axes(self):
        """ Create a list containing the three original x-, y- and z-axes
        and replace *None* with the amount of pixels along the given axis.
        """
        shapes = self.data.get_value().shape
        # Avoid undefined axes scales and replace them with len(1) sequences
        for i, axis in enumerate(self.axes):
            if axis is None:
                self.axes[i] = np.arange(shapes[i])

    def on_data_change(self):
        """ Update self.main_window.image_data and replot. """
        self.update_image_data()
        # self.main_window.redraw_plots()
        # Also need to recalculate the intensity plot
        # self.on_z_dim_change()

    def on_z_dim_change(self):
        """ Called when either completely new data is loaded or the dimension
        from which we look at the data changed (e.g. through :func:`roll_axes
        <data_slicer.pit.PITDataHandler.roll_axes>`).
        Update the z range and the integrated intensity plot.
        """

        # Calculate the integrated intensity and plot it
        self.calculate_integrated_intensity()
        # ip.plot(self.integrated)

    def calculate_integrated_intensity(self):
        self.integrated = self.get_data().sum(0).sum(0)

    def update_image_data(self):
        """ Get the right (possibly integrated) slice out of *self.data*,
        apply postprocessings and store it in *self.image_data*.
        Skip this if the z value happens to be out of range, which can happen
        if the image data changes and the z scale hasn't been updated yet.
        """
        # z = self.z.get_value()
        # integrate_z = self.main_window.plot_z.width
        data = self.get_data()
        try:
            self.main_window.image_data = data
        except IndexError:
            print(('update_image_data(): z index {} out of range for data of length {}.'))


class MainWindow2D(QtWidgets.QMainWindow):

    def __init__(self, data_browser, data_set=None, index=None, slice=False):
        super(MainWindow2D, self).__init__()
        self.title = index.split('/')[-1]
        self.fname = index
        self.central_widget = QtWidgets.QWidget()
        self.layout = QtWidgets.QGridLayout()

        # moved to get rid of warnings
        self.cmap = None
        self.image_kwargs = None
        self.cmap_name = None
        self.lut = None
        self.db = data_browser
        self.index = index
        self.slice = slice
        self.slider_pos = [0, 0]
        self.new_energy_axis = None
        self.k_axis = None
        self.smooth = False
        self.curvature = False
        self.thread = {}
        # self.thread_count = 0
        self.data_viewers = {}

        # Initialize instance variables
        # Plot transparency alpha
        self.alpha = 1
        # Plot powerlaw normalization exponent gamma
        self.gamma = 1
        # Relative colormap maximum
        self.vmax = 1

        # Need to store original transformation information for `rotate()`
        self._transform_factors = []

        self.data_set = deepcopy(data_set)
        self.org_dataset = None
        self.data_handler = DataHandler2D(self)

        # Create the 3D (main) and cut ImagePlots
        self.main_plot = ip.ImagePlot(name='main_plot', crosshair=True)
        # Create cut of cut_x
        self.plot_x = ip.CursorPlot(name='plot_x')
        # Create cut of cut_y
        self.plot_y = ip.CursorPlot(name='plot_y', orientation='vertical')
        # Create utilities panel
        self.util_panel = ip.UtilitiesPanel(self, name='utilities_panel', dim=2)
        self.util_panel.positions_momentum_label.setText('Sliders')
        # self.util_panel.momentum_vert_label.setText('E:')

        self.setStyleSheet(app_style)
        self.set_cmap()

        self.setGeometry(50, 50, 700, 800)
        self.setWindowTitle(self.title)
        self.initUI()

        time_dl_b = time.time()
        # Set the loaded data in PIT
        if data_set is None:
            print('Data set not defined.')
        else:
            raw_data = data_set

        D = self.swap_axes_aroud(raw_data)

        # print(f'data shape = {D.data.shape}')
        # print(f'x shape = {D.xscale.shape}')
        # print(f'y shape = {D.yscale.shape}')
        # print(f'z shape = {D.zscale.shape}')
        self.data_handler.prepare_data(D.data, [D.xscale, D.yscale, D.zscale])
        self.org_image_data = self.data_handler.get_data()

        self.util_panel.energy_vert.setRange(0, len(self.data_handler.axes[1]))
        self.util_panel.momentum_hor.setRange(0, len(self.data_handler.axes[0]))

        try:
            self.load_saved_corrections(data_set)
        except AttributeError:
            print('going with old settings.')
            self.load_saved_corrections_old(data_set)

        self.util_panel.set_metadata_window(raw_data)
        self.put_sliders_in_initial_positions()

    def initUI(self):
        self.setWindowTitle(self.title)
        # Create a "central widget" and its layout
        self.central_widget.setLayout(self.layout)
        self.setCentralWidget(self.central_widget)
        self.util_panel.bin_y_nbins.setValue(10)
        self.util_panel.bin_z_nbins.setValue(5)

        # Create plot_x
        self.plot_x.register_traced_variable(self.main_plot.pos[0])
        self.plot_x.pos.sig_value_changed.connect(self.update_plot_y)

        # Create plot_y
        self.plot_y.register_traced_variable(self.main_plot.pos[1])
        self.plot_y.pos.sig_value_changed.connect(self.update_plot_x)

        # Create utilities panel and connect signals
        self.util_panel.image_cmaps.currentIndexChanged.connect(self.set_cmap)
        self.util_panel.image_invert_colors.stateChanged.connect(self.set_cmap)
        self.util_panel.image_gamma.valueChanged.connect(self.set_gamma)
        self.util_panel.image_colorscale.valueChanged.connect(self.set_alpha)
        self.util_panel.image_normalize_edcs.stateChanged.connect(self.update_main_plot)
        self.util_panel.image_smooth_button.clicked.connect(self.smoooth_data)
        self.util_panel.image_curvature_button.clicked.connect(self.curvature_method)

        self.util_panel.file_mdc_fitter_button.clicked.connect(self.open_mdc_fitter)
        self.util_panel.file_edc_fitter_button.clicked.connect(self.open_edc_fitter)

        # binning utilities
        self.util_panel.bin_y.stateChanged.connect(self.update_binning_lines)
        self.util_panel.bin_y_nbins.valueChanged.connect(self.update_binning_lines)
        self.util_panel.bin_z.stateChanged.connect(self.update_binning_lines)
        self.util_panel.bin_z_nbins.valueChanged.connect(self.update_binning_lines)
        self.util_panel.energy_vert.valueChanged.connect(self.set_vert_energy_slider)
        self.util_panel.momentum_hor.valueChanged.connect(self.set_hor_momentum_slider)

        # buttons
        self.util_panel.close_button.clicked.connect(self.close)
        self.util_panel.save_button.clicked.connect(self.save_to_pickle)

        # energy and k-space concersion
        self.util_panel.axes_energy_Ef.valueChanged.connect(self.apply_energy_correction)
        self.util_panel.axes_energy_hv.valueChanged.connect(self.apply_energy_correction)
        self.util_panel.axes_energy_wf.valueChanged.connect(self.apply_energy_correction)
        self.util_panel.axes_energy_scale.currentIndexChanged.connect(self.apply_energy_correction)
        self.util_panel.axes_do_kspace_conv.clicked.connect(self.convert_to_kspace)
        self.util_panel.axes_reset_conv.clicked.connect(self.reset_kspace_conversion)

        # Align all the gui elements
        self._align()
        self.show()

    def _align(self):
        """ Align all the GUI elements in the QLayout::

              0   1   2   3
            +---+---+---+---+
            |utilities panel| 0
            +---+---+---+---+
            | mdc x |       | 1
            +-------+  edc  |
            | cut x |       | 2
            +-------+-------+
            |       | c | m | 3
            | main  | y | y | 4
            +---+---+---+---+

            (Units of subdivision [sd])
        """
        # subdivision
        sd = 1
        # Get a short handle
        l = self.layout
        # addWIdget(row, column, rowSpan, columnSpan)
        # utilities bar
        # l.addWidget(self.b0, 0, 0, 1 * sd, 5 * sd)
        l.addWidget(self.util_panel, 0, 0, 2 * sd, 6 * sd)
        # EDC
        l.addWidget(self.plot_x, 2 * sd, 0, 3 * sd, 4 * sd)
        # Main plot
        l.addWidget(self.main_plot, 5 * sd, 0, 6 * sd, 4 * sd)
        # MDC
        l.addWidget(self.plot_y, 5 * sd, 4 * sd, 6 * sd, 2 * sd)

        nrows = 12 * sd
        ncols = 6 * sd
        # Need to manually set all row- and columnspans as well as min-sizes
        for i in range(nrows):
            l.setRowMinimumHeight(i, 50)
            l.setRowStretch(i, 1)
        for i in range(ncols):
            l.setColumnMinimumWidth(i, 50)
            l.setColumnStretch(i, 1)

    def closeEvent(self, event) :
        """ Ensure that this instance is un-registered from the DataBrowser. """
        self.db.thread[self.index].quit()
        self.db.thread[self.index].wait()
        del(self.db.thread[self.index])
        del(self.db.data_viewers[self.index])

    @staticmethod
    def swap_axes_aroud(D):
        """
        Swap axes and data dimensions to fit rest of the code
        """
        tmp = deepcopy(D)
        D.xscale = tmp.yscale
        D.yscale = tmp.zscale
        D.zscale = tmp.xscale
        D.data = np.ones((1, tmp.zscale.size, tmp.yscale.size))
        D.data[0, :, :] = np.rollaxis(tmp.data[0, :, :], 1)
        # print('viewer format')
        # print(f'data shape: {D.data.shape}')
        # print(f'x shape: {D.xscale.shape}')
        # print(f'y shape: {D.yscale.shape}')
        # print(f'z shape: {D.zscale.shape}')

        return D

    def update_main_plot(self, **image_kwargs):
        """ Change *self.main_plot*`s currently displayed
        :class:`image_item <data_slicer.imageplot.ImagePlot.image_item>` to
        the slice of *self.data_handler.data* corresponding to the current
        value of *self.z*.
        """

        self.data_handler.update_image_data()

        if image_kwargs != {}:
            self.image_kwargs = image_kwargs
        # Add image to main_plot
        self.set_image(self.image_data, **image_kwargs)

    def set_axes(self):
        """ Set the x- and y-scales of the plots. The :class:`ImagePlot
        <data_slicer.imageplot.ImagePlot>` object takes care of keeping the
        scales as they are, once they are set.
        """
        xaxis = self.data_handler.axes[1]
        yaxis = self.data_handler.axes[0]
        self.main_plot.set_xscale(range(0, len(xaxis)))
        self.main_plot.set_ticks(xaxis[0], xaxis[-1], self.main_plot.main_xaxis)
        self.plot_x.set_ticks(xaxis[0], xaxis[-1], self.plot_x.main_xaxis)
        self.plot_x.set_secondary_axis(0, len(xaxis))
        self.main_plot.set_yscale(range(0, len(yaxis)))
        self.main_plot.set_ticks(yaxis[0], yaxis[-1], self.main_plot.main_yaxis)
        self.plot_y.set_ticks(yaxis[0], yaxis[-1], self.plot_y.main_xaxis)
        self.plot_y.set_secondary_axis(0, len(yaxis))
        self.main_plot.fix_viewrange()

    def update_plot_x(self):
        # Get shorthands for plot
        xp = self.plot_x
        try:
            old = xp.listDataItems()[0]
            xp.removeItem(old)
        except IndexError:
            pass

        # Get the correct position indicator
        pos = self.main_plot.pos[1]
        binning = self.util_panel.bin_y.isChecked()
        if binning:
            width = self.util_panel.bin_y_nbins.value()
        else:
            width = 0
        if pos.allowed_values is not None:
            i_x = int(min(pos.get_value(), pos.allowed_values.max() - 1))
        else:
            i_x = 0

        try:
            data = self.image_data
        except AttributeError:
            data = self.data_handler.get_data()

        if width == 0:
            y = data[:, i_x]
        else:
            start = i_x - width
            stop = i_x + width
            y = np.sum(data[:, start:stop], axis=1)
            # y = wp.normalize(y)
        x = np.arange(0, len(self.data_handler.axes[1]))
        self.edc = y
        xp.plot(x, y)
        self.util_panel.momentum_hor.setValue(i_x)
        if self.k_axis is None:
            self.util_panel.momentum_hor_value.setText('({:.4f})'.format(self.data_handler.axes[0][i_x]))
        else:
            self.util_panel.momentum_hor_value.setText('({:.4f})'.format(self.k_axis[i_x]))

        if binning:
            self.plot_y.left_line.setValue(i_x - width)
            self.plot_y.right_line.setValue(i_x + width)

    def update_plot_y(self):
        # Get shorthands for plot
        yp = self.plot_y
        try:
            old = yp.listDataItems()[0]
            yp.removeItem(old)
        except IndexError:
            pass

        # Get the correct position indicator
        pos = self.main_plot.pos[0]
        binning = self.util_panel.bin_z.isChecked()
        if binning:
            width = self.util_panel.bin_z_nbins.value()
        else:
            width = 0
        if pos.allowed_values is not None:
            i_y = int(min(pos.get_value(), pos.allowed_values.max() - 1))
        else:
            i_y = 0
        try:
            data = self.image_data
        except AttributeError:
            data = self.data_handler.get_data()
        if width == 0:
            y = data[i_y, :]
        else:
            start = i_y - width
            stop = i_y + width
            y = np.sum(data[start:stop, :], axis=0)
            # y = wp.normalize(y)
        x = np.arange(0, len(self.data_handler.axes[0]))
        self.mdc = y
        yp.plot(y, x)
        self.util_panel.energy_vert.setValue(i_y)

        if self.new_energy_axis is None:
            self.util_panel.energy_vert_value.setText('({:.4f})'.format(self.data_handler.axes[1][i_y]))
        else:
            self.util_panel.energy_vert_value.setText('({:.4f})'.format(self.new_energy_axis[i_y]))

        if binning:
            self.plot_x.left_line.setValue(i_y - width)
            self.plot_x.right_line.setValue(i_y + width)

    def update_binning_lines(self):
        """ Update binning lines accordingly. """
        # edc plot
        if self.util_panel.bin_y.isChecked():
            try:
                half_width = self.util_panel.bin_y_nbins.value()
                pos = self.main_plot.pos[1].get_value()
                self.main_plot.add_binning_lines(pos, half_width)
                self.plot_y.add_binning_lines(pos, half_width)
                ymin = 0 + half_width
                ymax = len(self.data_handler.axes[0]) - half_width
                new_range = np.arange(ymin, ymax)
                self.main_plot.pos[1].set_allowed_values(new_range)
                self.plot_y.pos.set_allowed_values(new_range)
            except AttributeError:
                pass
        else:
            try:
                self.main_plot.remove_binning_lines()
                self.plot_y.remove_binning_lines()
                org_range = np.arange(0, len(self.data_handler.axes[0]))
                self.main_plot.pos[1].set_allowed_values(org_range)
                self.plot_y.pos.set_allowed_values(org_range)
            except AttributeError:
                pass

        # mdc plot
        if self.util_panel.bin_z.isChecked():
            try:
                half_width = self.util_panel.bin_z_nbins.value()
                pos = self.main_plot.pos[0].get_value()
                self.main_plot.add_binning_lines(pos, half_width, orientation='vertical')
                self.plot_x.add_binning_lines(pos, half_width)
                ymin = 0 + half_width
                ymax = len(self.data_handler.axes[1]) - half_width
                new_range = np.arange(ymin, ymax)
                self.main_plot.pos[0].set_allowed_values(new_range)
                self.plot_x.pos.set_allowed_values(new_range)
            except AttributeError:
                pass
        else:
            try:
                self.main_plot.remove_binning_lines(orientation='vertical')
                self.plot_x.remove_binning_lines()
                org_range = np.arange(0, len(self.data_handler.axes[1]))
                self.main_plot.pos[0].set_allowed_values(org_range)
                self.plot_x.pos.set_allowed_values(org_range)
            except AttributeError:
                pass

    def set_cmap(self):
        """ Set the colormap to *cmap* where *cmap* is one of the names
        registered in :mod:`<data_slicer.cmaps>` which includes all matplotlib and
        kustom cmaps.
        WP: small changes made to use only my list of cmaps (see cmaps.py) and to shorten the list
        by using 'invert_colors' checkBox
        """
        try:
            cmap = self.util_panel.image_cmaps.currentText()
            if self.util_panel.image_invert_colors.isChecked() and MY_CMAPS:
                cmap = cmap + '_r'
        except AttributeError:
            cmap = DEFAULT_CMAP

        try:
            self.cmap = cmaps[cmap]
        except KeyError:
            print('Invalid colormap name. Use one of: ')
            print(cmaps.keys())
        self.cmap_name = cmap
        # Since the cmap changed it forgot our settings for alpha and gamma
        self.cmap.set_alpha(self.alpha)
        self.cmap.set_gamma()
        sliders_pos = self.get_sliders_positions()
        self.cmap_changed()
        self.set_sliders_postions(sliders_pos)
        self.update_binning_lines()

    def cmap_changed(self):
        """ Recalculate the lookup table and redraw the plots such that the
        changes are immediately reflected.
        """
        self.lut = self.cmap.getLookupTable()
        self.redraw_plots()

    def redraw_plots(self, image=None):
        """ Redraw plotted data to reflect changes in data or its colors. """
        try:
            # Redraw main plot
            self.set_image(image, displayed_axes=self.data_handler.displayed_axes)
            # Redraw cut plot
            # self.update_cut()
        except AttributeError as e:
            # In some cases (namely initialization) the mainwindow is not defined yet
            pass
            # print('AttributeError: {}'.format(e))

    def set_image(self, image=None, *args, **kwargs):
        """ Wraps the underlying ImagePlot3d's set_image method.
        See :func:`~data_slicer.imageplot.ImagePlot3d.set_image`. *image* can
        be *None* i.e. in order to just update the plot with a new colormap.
        """

        # Reset the transformation
        self._transform_factors = []
        if image is None:
            image = self.image_data
        self.main_plot.set_image(image, *args, lut=self.lut, **kwargs)

    def set_alpha(self, alpha):
        """ Set the alpha value of the currently used cmap. *alpha* can be a
        single float or an array of length ``len(self.cmap.color)``.
        """
        self.alpha = alpha
        sliders_pos = self.get_sliders_positions()
        self.cmap.set_alpha(alpha)
        self.cmap_changed()
        self.set_sliders_postions(sliders_pos)
        self.update_binning_lines()

    def set_gamma(self):
        """ Set the exponent for the power-law norm that maps the colors to
        values. I.e. the values where the colours are defined are mapped like
        ``y=x**gamma``.
        WP: changed to work with applied QDoubleSpinBox
        """
        gamma = self.util_panel.image_gamma.value()
        self.gamma = gamma
        sliders_pos = self.get_sliders_positions()
        self.cmap.set_gamma(gamma)
        self.cmap_changed()
        self.set_sliders_postions(sliders_pos)
        self.update_binning_lines()

    def set_colorscale(self):
        """ Set the relative maximum of the colormap. I.e. the colors are
        mapped to the range `min(data)` - `vmax*max(data)`.
        WP: changed to work with applied QDoubleSpinBox
        """
        vmax = self.util_panel.image_colorscale.value()
        self.vmax = vmax
        self.cmap.set_vmax(vmax)
        self.cmap_changed()

    def set_vert_energy_slider(self):
        energy = self.util_panel.energy_vert.value()
        self.main_plot.pos[0].set_value(energy)
        self.update_binning_lines()

    def set_hor_momentum_slider(self):
        angle = self.util_panel.momentum_hor.value()
        self.main_plot.pos[1].set_value(angle)
        self.update_binning_lines()

    def set_sliders_postions(self, positions):
        self.main_plot.pos[1].set_value(positions[0])
        self.main_plot.pos[0].set_value(positions[1])

    def get_sliders_positions(self):
        main_hor = self.main_plot.crosshair.hpos.get_value()
        main_ver = self.main_plot.crosshair.vpos.get_value()
        return [main_hor, main_ver]

    def put_sliders_in_initial_positions(self):
        if self.new_energy_axis is None:
            e_ax = self.data_handler.axes[1]
        else:
            e_ax = self.new_energy_axis
        if (e_ax.min() < 0) and (e_ax.max() > 0):
            mid_energy = wp.indexof(-0.005, e_ax)
        else:
            mid_energy = int(len(e_ax) / 2)
            self.util_panel.axes_energy_scale.setCurrentIndex(1)

        if self.k_axis is None:
            mom_ax = self.data_handler.axes[0]
        else:
            mom_ax = self.k_axis
        if (mom_ax.min() < 0) and (mom_ax.max() > 0):
            mid_angle = wp.indexof(0, mom_ax)
        else:
            mid_angle = int(len(mom_ax) / 2)

        self.main_plot.pos[0].set_value(mid_energy)
        self.main_plot.pos[1].set_value(mid_angle)

    def apply_energy_correction(self):
        Ef = self.util_panel.axes_energy_Ef.value()
        hv = self.util_panel.axes_energy_hv.value()
        wf = self.util_panel.axes_energy_wf.value()

        scale = self.util_panel.axes_energy_scale.currentText()
        if scale == 'binding':
            hv = 0
            wf = 0

        new_energy_axis = self.data_handler.axes[1] + Ef - hv + wf
        self.new_energy_axis = new_energy_axis
        new_range = [new_energy_axis[0], new_energy_axis[-1]]
        self.main_plot.plotItem.getAxis(self.main_plot.main_xaxis).setRange(*new_range)
        self.plot_x.plotItem.getAxis(self.plot_x.main_xaxis).setRange(*new_range)

        # update energy labels
        erg_idx = self.main_plot.crosshair.vpos.get_value()
        self.util_panel.momentum_hor_value.setText('({:.4f})'.format(self.new_energy_axis[erg_idx]))

    def convert_to_kspace(self):
        scan_ax = np.array([0])
        anal_axis = self.data_handler.axes[0]
        d_scan_ax = self.util_panel.axes_angle_off.value()
        d_anal_ax = self.data_handler.axes[0][self.util_panel.axes_gamma_x.value()]
        orientation = self.util_panel.axes_slit_orient.currentText()
        a = self.util_panel.axes_conv_lc.value()
        energy = self.new_energy_axis
        hv = self.util_panel.axes_energy_hv.value()
        wf = self.util_panel.axes_energy_wf.value()

        if hv == 0 or wf == 0:
            warning_box = QMessageBox()
            warning_box.setIcon(QMessageBox.Information)
            warning_box.setWindowTitle('Wrong conversion values.')
            if hv == 0 and wf == 0:
                msg = 'Photon energy and work fonction values not given.'
            elif hv == 0:
                msg = 'Photon energy value not given.'
            elif wf == 0:
                msg = 'Work fonction value not given.'
            warning_box.setText(msg)
            warning_box.setStandardButtons(QMessageBox.Ok)
            if warning_box.exec() == QMessageBox.Ok:
                return

        nma, erg = wp.angle2kscape(scan_ax, anal_axis, d_scan_ax=d_scan_ax, d_anal_ax=d_anal_ax,
                                   orientation=orientation, a=a, energy=energy, hv=hv, work_func=wf)
        self.k_axis = nma[-1]
        new_range = [nma[-1][0], nma[-1][-1]]
        self.main_plot.plotItem.getAxis(self.main_plot.main_yaxis).setRange(*new_range)
        self.plot_y.plotItem.getAxis(self.plot_y.main_xaxis).setRange(*new_range)
        self.util_panel.momentum_hor_value.setText('({:.4f})'.format(self.k_axis[self.main_plot.pos[1].get_value()]))

    def reset_kspace_conversion(self):
        self.k_axis = None
        org_range = [self.data_handler.axes[0][0], self.data_handler.axes[0][-1]]
        self.main_plot.plotItem.getAxis(self.main_plot.main_yaxis).setRange(*org_range)
        self.plot_y.plotItem.getAxis(self.plot_y.main_xaxis).setRange(*org_range)
        self.util_panel.momentum_hor_value.setText('({:.4f})'.format(
            self.data_handler.axes[0][self.main_plot.pos[1].get_value()]))

    def smoooth_data(self):
        self.smooth = not self.smooth
        if self.smooth:
            data = self.data_handler.get_data()
            nb = self.util_panel.image_smooth_n.value()
            rl = self.util_panel.image_smooth_rl.value()
            self.image_data = wp.smooth_2d(data, n_box=nb, recursion_level=rl)
            self.set_image()
            self.util_panel.image_smooth_button.setText('Reset')
        else:
            self.image_data = deepcopy(self.org_image_data)
            self.set_image()
            self.curvature = False
            self.util_panel.image_smooth_button.setText('Smooth')
            self.util_panel.image_curvature_button.setText('Do curvature')

    def curvature_method(self):
        self.curvature = not self.curvature
        if self.curvature:
            a = self.util_panel.image_curvature_a.value()
            if self.k_axis is None:
                dx = wp.get_step(self.data_handler.axes[0])
            else:
                dx = wp.get_step(self.k_axis)
            if self.new_energy_axis is None:
                dy = wp.get_step(self.data_handler.axes[1])
            else:
                dy = wp.get_step(self.new_energy_axis)
            self.smooth_data = deepcopy(self.image_data)
            method = self.util_panel.image_curvature_method.currentText()
            if method == '2D':
                self.image_data = wp.curvature_2d(self.image_data, dx, dy, a0=a)
            elif method == '1D (EDC)':
                for idx in range(self.image_data.shape[1]):
                    self.image_data[:, idx] = wp.curvature_1d(self.image_data[:, idx], a0=a)
            elif method == '1D (MDC)':
                for idx in range(self.image_data.shape[0]):
                    self.image_data[idx, :] = wp.curvature_1d(self.image_data[idx, :], a0=a)
            self.set_image()
            self.util_panel.image_curvature_button.setText('Reset')
        else:
            if self.smooth:
                self.image_data = self.smooth_data
            else:
                self.image_data = deepcopy(self.org_image_data)
            self.set_image()
            self.util_panel.image_curvature_button.setText('Do curvature')

    def open_mdc_fitter(self):

        thread_lbl = self.index + '_mdc_viewer'
        self.mdc_thread_lbl = thread_lbl

        if thread_lbl in self.data_viewers:
            already_opened_box = QMessageBox()
            already_opened_box.setIcon(QMessageBox.Information)
            already_opened_box.setText('MDC viewer already opened.')
            already_opened_box.setStandardButtons(QMessageBox.Ok)
            if already_opened_box.exec() == QMessageBox.Ok:
                return

        self.thread[thread_lbl] = ip.ThreadClass(index=thread_lbl)
        self.thread[thread_lbl].start()
        title = self.title + ' - mdc fitter'
        if self.new_energy_axis is not None:
            erg_ax = self.new_energy_axis
        else:
            erg_ax = self.data_set.zscale
        if self.k_axis is not None:
            k_ax = self.k_axis
        else:
            k_ax = self.data_set.yscale
        axes = [k_ax, erg_ax]
        try:
            self.data_viewers[thread_lbl] = MDCFitter(self, self.data_set, axes, title, index=thread_lbl)
        except Exception:
            error_box = QMessageBox()
            error_box.setIcon(QMessageBox.Information)
            error_box.setText('Couldn\'t load data,  something went wrong.')
            error_box.setStandardButtons(QMessageBox.Ok)
            if error_box.exec() == QMessageBox.Ok:
                return

    def open_edc_fitter(self):

        thread_idx = self.index + '_edc_viewer'
        self.edc_thread_lbl = thread_idx

        if thread_idx in self.data_viewers:
            already_opened_box = QMessageBox()
            already_opened_box.setIcon(QMessageBox.Information)
            already_opened_box.setText('EDC viewer already opened.')
            already_opened_box.setStandardButtons(QMessageBox.Ok)
            if already_opened_box.exec() == QMessageBox.Ok:
                return

        self.thread[thread_idx] = ip.ThreadClass(index=thread_idx)
        self.thread[thread_idx].start()
        title = self.title + ' - edc fitter'
        if self.new_energy_axis is not None:
            erg_ax = self.new_energy_axis
        else:
            erg_ax = self.data_set.zscale
        if self.k_axis is not None:
            k_ax = self.k_axis
        else:
            k_ax = self.data_set.yscale
        axes = [k_ax, erg_ax]
        try:
            self.data_viewers[thread_idx] = EDCFitter(self, self.data_set, axes, title, index=thread_idx)
        except Exception:
            error_box = QMessageBox()
            error_box.setIcon(QMessageBox.Information)
            error_box.setText('Couldn\'t load data,  something went wrong.')
            error_box.setStandardButtons(QMessageBox.Ok)
            if error_box.exec() == QMessageBox.Ok:
                return

    def save_to_pickle(self):
        # TODO change 'k_axis' to 'k' for all cuts: add 'change attrs name'
        # energy axis
        dataset = self.data_set
        dir = self.fname[:-len(self.title)]
        up = self.util_panel
        file_selection = True
        init_fname = self.title

        while file_selection:
            fname, fname_return_value = \
                    QtWidgets.QInputDialog.getText(self, '', 'File name:', 
                                                   QtWidgets.QLineEdit.Normal,
                                                   init_fname)
            if not fname_return_value:
                return

            # check if there is no fname colosions
            if fname in os.listdir(dir):
                fname_colision_box = QMessageBox()
                fname_colision_box.setIcon(QMessageBox.Question)
                fname_colision_box.setWindowTitle('File name already used.')
                fname_colision_box.setText('File {} already exists.\nDo you want to overwrite it?'.format(fname))
                fname_colision_box.setStandardButtons(QMessageBox.Ok | QMessageBox.Cancel)
                if fname_colision_box.exec() == QMessageBox.Ok:
                    file_selection = False
                else:
                    init_fname = fname
            else:
                file_selection = False

        conditions = [up.axes_energy_Ef.value() != 0, up.axes_energy_hv.value() != 0, up.axes_energy_wf.value() != 0,
                      up.axes_angle_off.value() != 0, up.axes_gamma_x.value() != 0]

        if np.any(conditions):
            save_cor_box = QMessageBox()
            save_cor_box.setIcon(QMessageBox.Question)
            save_cor_box.setWindowTitle('Save data')
            save_cor_box.setText("Do you want to save applied corrections?")
            save_cor_box.setStandardButtons(QMessageBox.No | QMessageBox.Ok | QMessageBox.Cancel)

            box_return_value = save_cor_box.exec()
            if box_return_value == QMessageBox.Ok:
                if up.axes_energy_Ef.value() != 0:
                    dataset.Ef = up.axes_energy_Ef.value()
                if up.axes_energy_hv.value() != 0:
                    dataset.hv = up.axes_energy_hv.value()
                if up.axes_energy_wf.value() != 0:
                    dataset.wf = up.axes_energy_wf.value()
                # if up.axes_angle_off.value() != 0:
                #     attrs['angle_off'] = up.axes_angle_off.value()
                if not (self.k_axis is None):
                    dataset.kyscale = self.k_axis
            elif box_return_value == QMessageBox.No:
                pass
            elif box_return_value == QMessageBox.Cancel:
                return
        else:
            pass

        dl.dump(dataset, (dir + fname), force=True)

    def load_saved_corrections(self, data_set):

        if hasattr(data_set, 'saved'):
            raise AttributeError
        if not (data_set.Ef is None):
            self.util_panel.axes_energy_Ef.setValue(data_set.Ef)
        if not (data_set.hv is None):
            self.util_panel.axes_energy_hv.setValue(data_set.hv)
        if not (data_set.wf is None):
            self.util_panel.axes_energy_wf.setValue(data_set.wf)
        if not (data_set.kyscale is None):
            self.k_axis = data_set.kyscale
            new_range = [self.k_axis[0], self.k_axis[-1]]
            self.main_plot.plotItem.getAxis(self.main_plot.main_yaxis).setRange(*new_range)
            self.plot_y.plotItem.getAxis(self.plot_y.main_xaxis).setRange(*new_range)
            self.util_panel.momentum_hor_value.setText('({:.4f})'.format(
                self.k_axis[self.main_plot.pos[1].get_value()]))

    def load_saved_corrections_old(self, data_set):

        if hasattr(data_set, 'saved'):
            saved = data_set.saved
            if 'Ef' in saved.keys():
                self.util_panel.axes_energy_Ef.setValue(saved['Ef'])
            if 'hv' in saved.keys():
                self.util_panel.axes_energy_hv.setValue(saved['hv'])
            if 'wf' in saved.keys():
                self.util_panel.axes_energy_wf.setValue(saved['wf'])
            if 'angle_off' in saved.keys():
                self.util_panel.axes_angle_off.setValue(saved['angle_off'])
            if 'k' in saved.keys():
                self.k_axis = saved['k']
                new_range = [self.k_axis[0], self.k_axis[-1]]
                self.main_plot.plotItem.getAxis(self.main_plot.main_yaxis).setRange(*new_range)
                self.plot_y.plotItem.getAxis(self.plot_y.main_xaxis).setRange(*new_range)
                self.util_panel.momentum_hor_value.setText('({:.4f})'.format(
                    self.k_axis[self.main_plot.pos[1].get_value()]))
            if 'k_axis' in saved.keys():
                self.k_axis = saved['k_axis']
                new_range = [self.k_axis[0], self.k_axis[-1]]
                self.main_plot.plotItem.getAxis(self.main_plot.main_yaxis).setRange(*new_range)
                self.plot_y.plotItem.getAxis(self.plot_y.main_xaxis).setRange(*new_range)
                self.util_panel.momentum_hor_value.setText('({:.4f})'.format(
                    self.k_axis[self.main_plot.pos[1].get_value()]))
        else:
            pass

        if not (data_set.Ef is None):
            self.util_panel.axes_energy_Ef.setValue(float(data_set.Ef))
        if not (data_set.hv is None):
            self.util_panel.axes_energy_hv.setValue(float(data_set.hv))
        if not (data_set.wf is None):
            self.util_panel.axes_energy_wf.setValue(float(data_set.wf))




import os

import numpy as np
from PyQt5.QtWidgets import QTabWidget, QWidget, QLabel, QCheckBox, QComboBox, QDoubleSpinBox, QSpinBox, QPushButton, \
    QLineEdit, QMainWindow, QMessageBox
from PyQt5.QtGui import QFont
from PyQt5 import QtCore
from pyqtgraph.Qt import QtGui, QtWidgets
from pyqtgraph import InfiniteLine, PlotWidget, AxisItem, mkPen, PColorMeshItem, mkBrush, FillBetweenItem, \
    PlotDataItem, ScatterPlotItem
from pyqtgraph.graphicsItems.ImageItem import ImageItem
from scipy.optimize import curve_fit
from scipy.optimize import OptimizeWarning

import piva.arpys_wp as wp
from piva.cmaps import cmaps, my_cmaps
from piva.imageplot import TracedVariable

BASE_LINECOLOR = (255, 255, 0, 255)
BINLINES_LINECOLOR = (168, 168, 104, 255)
ORIENTLINES_LINECOLOR = (164, 37, 22, 255)
HOVER_COLOR = (195, 155, 0, 255)
BGR_COLOR = (64, 64, 64)
MDC_PANEL_BGR = (236, 236, 236)
util_panel_style = """
QFrame{margin:5px; border:1px solid rgb(150,150,150);}
QLabel{color: rgb(246, 246, 246); border:1px solid rgb(64, 64, 64);}
QCheckBox{color: rgb(246, 246, 246);}
"""
SIGNALS = 5
MY_CMAPS = True
DEFAULT_CMAP = 'coolwarm'

bold_font = QFont()
bold_font.setBold(True)


class MDCFitter(QMainWindow):
    # TODO Smoothing

    def __init__(self, data_viewer, data_set, axes, title, index=None):
        super(MDCFitter, self).__init__()

        self.central_widget = QWidget()
        self.mdc_fitter_layout = QtWidgets.QGridLayout()
        self.central_widget.setLayout(self.mdc_fitter_layout)
        self.tabs = QTabWidget()

        self.settings_panel = QWidget()
        self.cut_panel = PlotWidget(background=MDC_PANEL_BGR)
        self.mdc_panel = PlotWidget(background=MDC_PANEL_BGR)

        self.data_viewer = data_viewer
        self.data_set = data_set
        self.title = title
        self.thread_index = index
        self.k_ax = axes[0]
        self.erg_ax = axes[1]
        self.fit_results = None

        self.align()
        self.set_settings_panel()
        self.set_cut_panel(data_set.data[0, :, :])
        self.set_mdc_panel()

        self.initUI()
        self.set_cmap()
        self.setCentralWidget(self.central_widget)
        self.setWindowTitle(self.title)
        self.show()

    def initUI(self):

        self.image_cmaps.currentIndexChanged.connect(self.set_cmap)
        self.image_invert_colors.stateChanged.connect(self.set_cmap)
        self.image_gamma.valueChanged.connect(self.set_gamma)
        self.image_e_pos.valueChanged.connect(self.update_mdc_slider)
        self.image_bin.stateChanged.connect(self.set_binning_lines)
        self.image_bin_n.valueChanged.connect(self.set_binning_lines)
        self.image_close_button.clicked.connect(self.close)

        self.fitting_button.clicked.connect(self.fit_mdc)
        self.fitting_range_start.valueChanged.connect(self.set_mdc_panel)
        self.fitting_range_stop.valueChanged.connect(self.set_mdc_panel)
        self.fitting_bgr_range_first.valueChanged.connect(self.set_mdc_panel)
        self.fitting_bgr_range_second.valueChanged.connect(self.set_mdc_panel)
        self.fitting_bgr_poly_button.clicked.connect(self.fit_bgr)

        self.fitting_result_append.clicked.connect(self.append_fit_results)
        self.fitting_result_edit.clicked.connect(self.edit_fit_results)
        self.fitting_result_update.clicked.connect(self.load_update_fit_result)
        self.fitting_result_save.clicked.connect(self.save_fit_results)

    def align(self):

        mdl = self.mdc_fitter_layout

        mdl.addWidget(self.tabs,             0, 0, 3, 8)
        mdl.addWidget(self.cut_panel,        4, 0, 4, 4)
        mdl.addWidget(self.mdc_panel,        4, 4, 4, 4)

    def set_settings_panel(self):

        self.set_image_tab()
        self.set_fitting_tab()

    def set_image_tab(self):
        # create elements
        self.image_tab = QWidget()
        itl = QtWidgets.QGridLayout()

        self.image_cmaps_label = QLabel('cmaps:')
        self.image_cmaps = QComboBox()
        self.image_invert_colors = QCheckBox('invert colors')
        self.image_gamma_label = QLabel('gamma:')
        self.image_gamma = QDoubleSpinBox()
        self.image_gamma.setRange(0.05, 10)
        self.image_gamma.setSingleStep(0.05)
        self.image_gamma.setValue(1)

        self.image_e_pos_lbl = QLabel('Energy:')
        self.image_e_pos = QSpinBox()
        self.image_e_pos.setRange(0, self.erg_ax.size)
        self.image_e_pos_value_lbl = QLabel('')

        self.image_k_pos_lbl = QLabel('Momentum:')
        self.image_k_pos = QSpinBox()
        self.image_k_pos.setRange(0, self.k_ax.size)
        self.image_k_pos_value_lbl = QLabel('')

        self.image_bin = QCheckBox('bin k')
        self.image_bin_n = QSpinBox()
        self.image_bin_n.setValue(3)

        self.image_close_button = QPushButton('close')

        row = 0
        itl.addWidget(self.image_e_pos_lbl,         row, 0)
        itl.addWidget(self.image_e_pos,             row, 1)
        itl.addWidget(self.image_e_pos_value_lbl,   row, 2, 1, 2)
        itl.addWidget(self.image_bin,               row, 4)
        itl.addWidget(self.image_bin_n,             row, 5)
        itl.addWidget(self.image_close_button,      row, 7)

        row = 1
        itl.addWidget(self.image_k_pos_lbl,         row, 0)
        itl.addWidget(self.image_k_pos,             row, 1)
        itl.addWidget(self.image_k_pos_value_lbl,   row, 2, 1, 2)

        row = 2
        itl.addWidget(self.image_cmaps_label,       row, 0)
        itl.addWidget(self.image_cmaps,             row, 1, 1, 2)
        itl.addWidget(self.image_invert_colors,     row, 3)
        itl.addWidget(self.image_gamma_label,       row, 4)
        itl.addWidget(self.image_gamma,             row, 5)

        self.setup_cmaps()
        self.image_tab.layout = itl
        self.image_tab.setLayout(itl)
        self.tabs.addTab(self.image_tab, 'Cut')

    def set_fitting_tab(self):
        self.fitting_tab = QWidget()
        ftl = QtWidgets.QGridLayout()

        self.fitting_mu_lbl = QLabel('\u03BC:')
        self.fitting_mu = QDoubleSpinBox()
        self.fitting_mu.setRange(self.k_ax.min(), self.k_ax.max())
        self.fitting_mu.setDecimals(4)
        self.fitting_gamma_lbl = QLabel('\u0393:')
        self.fitting_gamma = QDoubleSpinBox()
        self.fitting_gamma.setValue(0.1)
        self.fitting_mu.setDecimals(4)
        self.fitting_alpha_lbl = QLabel('\u03B1:')
        self.fitting_alpha = QDoubleSpinBox()
        self.fitting_alpha.setValue(1.)
        self.fitting_beta_lbl = QLabel('\u03B2:')
        self.fitting_beta = QDoubleSpinBox()
        self.fitting_beta.setValue(1.)
        self.fitting_a_lbl = QLabel('a:')
        self.fitting_a = QDoubleSpinBox()
        self.fitting_a.setValue(100)
        self.fitting_a.setRange(0, 1e5)
        self.fitting_button = QPushButton('fit MDC')

        self.fitting_message_cell = QLineEdit('Some stuff will show up here.')

        self.fitting_range_lbl = QLabel('fit range:')
        self.fitting_range_lbl.setFont(bold_font)
        # self.fitting_range_start = QSlider(QtCore.Qt.Orientation.Horizontal)
        self.fitting_range_start = QDoubleSpinBox()
        self.fitting_range_start.setRange(self.k_ax.min(), self.k_ax.max())
        self.fitting_range_start.setSingleStep(wp.get_step(self.k_ax))
        self.fitting_range_start.setDecimals(6)
        # self.fitting_range_start.setValue(self.k_ax.min())
        self.fitting_range_start.setValue(-0.84)

        self.fitting_range_stop = QDoubleSpinBox()
        self.fitting_range_stop.setRange(self.k_ax.min(), self.k_ax.max())
        self.fitting_range_stop.setSingleStep(wp.get_step(self.k_ax))
        self.fitting_range_stop.setDecimals(6)
        # self.fitting_range_stop.setValue(self.k_ax.max())
        self.fitting_range_stop.setValue(-0.3)

        self.fitting_bgr_range_lbl = QLabel('bgr range:')
        self.fitting_bgr_range_lbl.setFont(bold_font)
        self.fitting_bgr_range_first = QDoubleSpinBox()
        self.fitting_bgr_range_first.setRange(self.k_ax.min(), self.k_ax.max())
        self.fitting_bgr_range_first.setSingleStep(wp.get_step(self.k_ax))
        self.fitting_bgr_range_first.setDecimals(6)
        # self.fitting_bgr_range_first.setValue(self.k_ax.min())
        self.fitting_bgr_range_first.setValue(-0.66)

        self.fitting_bgr_range_second = QDoubleSpinBox()
        self.fitting_bgr_range_second.setRange(self.k_ax.min(), self.k_ax.max())
        self.fitting_bgr_range_second.setSingleStep(wp.get_step(self.k_ax))
        self.fitting_bgr_range_second.setDecimals(6)
        # self.fitting_bgr_range_second.setValue(self.k_ax.max())
        self.fitting_bgr_range_second.setValue(-0.4)

        self.fitting_bgr_poly_lbl = QLabel('bgr poly:')
        self.fitting_bgr_poly_lbl.setFont(bold_font)
        self.fitting_bgr_poly_order_lbl = QLabel('order')
        self.fitting_bgr_poly_order = QSpinBox()
        self.fitting_bgr_poly_order.setRange(1, 10)
        self.fitting_bgr_poly_button = QPushButton('fit bgr')

        self.fitting_results_lbl = QLabel('results:')
        self.fitting_results_lbl.setFont(bold_font)
        self.fitting_result_append = QPushButton('append')
        self.fitting_result_edit = QPushButton('edit')
        self.fitting_result_update = QPushButton('load/update')
        self.fitting_result_save = QPushButton('save')

        row = 0
        ftl.addWidget(self.fitting_mu_lbl,              row, 0)
        ftl.addWidget(self.fitting_mu,                  row, 1)
        ftl.addWidget(self.fitting_gamma_lbl,           row, 2)
        ftl.addWidget(self.fitting_gamma,               row, 3)
        ftl.addWidget(self.fitting_range_lbl,           row, 5)
        ftl.addWidget(self.fitting_range_start,         row, 6, 1, 2)
        ftl.addWidget(self.fitting_range_stop,          row, 8, 1, 2)

        row = 1
        ftl.addWidget(self.fitting_alpha_lbl,           row, 0)
        ftl.addWidget(self.fitting_alpha,               row, 1)
        ftl.addWidget(self.fitting_beta_lbl,            row, 2)
        ftl.addWidget(self.fitting_beta,                row, 3)
        ftl.addWidget(self.fitting_bgr_range_lbl,       row, 5)
        ftl.addWidget(self.fitting_bgr_range_first,     row, 6, 1, 2)
        ftl.addWidget(self.fitting_bgr_range_second,    row, 8, 1, 2)

        row = 2
        ftl.addWidget(self.fitting_a_lbl,               row, 0)
        ftl.addWidget(self.fitting_a,                   row, 1)
        ftl.addWidget(self.fitting_button,              row, 2, 1, 2)
        ftl.addWidget(self.fitting_bgr_poly_lbl,        row, 5)
        ftl.addWidget(self.fitting_bgr_poly_order_lbl,  row, 6)
        ftl.addWidget(self.fitting_bgr_poly_order,      row, 7)
        ftl.addWidget(self.fitting_bgr_poly_button,     row, 8, 1, 2)

        row = 3
        ftl.addWidget(self.fitting_message_cell,        row, 0, 1, 5)
        ftl.addWidget(self.fitting_results_lbl,         row, 5)
        ftl.addWidget(self.fitting_result_append,       row, 6)
        ftl.addWidget(self.fitting_result_edit,         row, 7)
        ftl.addWidget(self.fitting_result_update,       row, 8)
        ftl.addWidget(self.fitting_result_save,         row, 9)
        #
        # dummy_lbl = QLabel('')
        # ftl.addWidget(dummy_lbl, 2, 0, 1, 8)

        self.fitting_tab.layout = ftl
        self.fitting_tab.setLayout(ftl)
        self.tabs.addTab(self.fitting_tab, 'Fitting')

    def setup_cmaps(self):

        cm = self.image_cmaps
        if MY_CMAPS:
            cm.addItems(my_cmaps)
        else:
            for cmap in cmaps.keys():
                cm.addItem(cmap)
        cm.setCurrentText(DEFAULT_CMAP)

    def set_cmap(self):
        """ Set the colormap to *cmap* where *cmap* is one of the names
        registered in :mod:`<data_slicer.cmaps>` which includes all matplotlib and
        kustom cmaps.
        WP: small changes made to use only my list of cmaps (see cmaps.py) and to shorten the list
        by using 'invert_colors' checkBox
        """
        try:
            cmap = self.image_cmaps.currentText()
            if self.image_invert_colors.isChecked() and MY_CMAPS:
                cmap = cmap + '_r'
        except AttributeError:
            cmap = DEFAULT_CMAP

        try:
            self.cmap = cmaps[cmap]
        except KeyError:
            print('Invalid colormap name. Use one of: ')
            print(cmaps.keys())
        self.cmap_name = cmap
        # Since the cmap changed it forgot our settings for alpha and gamma
        self.cmap.set_alpha(1)
        self.cmap.set_gamma()
        self.cmap_changed()

    def cmap_changed(self, image=None, *args, **kwargs):
        """ Recalculate the lookup table and redraw the plots such that the
        changes are immediately reflected.
        """
        self.lut = self.cmap.getLookupTable()
        self._transform_factors = []
        if image is None:
            image = self.data_set.data[0, :, :]
        image_item = ImageItem(image, *args, lut=self.lut, **kwargs)

        self.cut_panel.removeItem(self.image_item)
        self.cut_panel.removeItem(self.mdc_line)
        self.image_item = image_item
        self.image_data = image
        self.cut_panel.addItem(image_item)
        self.cut_panel.addItem(self.mdc_line)
        self.set_edc_line()
        self.set_binning_lines()

    def set_gamma(self):
        """ Set the exponent for the power-law norm that maps the colors to
        values. I.e. the values where the colours are defined are mapped like
        ``y=x**gamma``.
        WP: changed to work with applied QDoubleSpinBox
        """
        gamma = self.image_gamma.value()
        self.gamma = gamma
        self.cmap.set_gamma(gamma)
        self.cmap_changed()

    def set_cut_panel(self, image):

        # Show top and tight axes by default, but without ticklabels
        self.cut_panel.showAxis('top')
        self.cut_panel.showAxis('right')
        self.cut_panel.getAxis('top').setStyle(showValues=False)
        self.cut_panel.getAxis('right').setStyle(showValues=False)
        self.cut_panel.main_xaxis = 'bottom'
        self.cut_panel.main_xaxis_grid = (255, 1)
        self.cut_panel.main_yaxis = 'left'
        self.cut_panel.main_yaxis_grid = (2, 0)

        # moved here to get rid of warnings:
        self.cut_panel.right_axis = 'top'
        self.cut_panel.angle = 90
        self.cut_panel.slider_axis_index = 1

        # Convert array to ImageItem
        if isinstance(image, np.ndarray):
            if 0 not in image.shape:
                self.image_item = ImageItem(image)
            else:
                return
        else:
            self.image_item = image

        self.cut_panel.addItem(self.image_item)
        self.set_ticks(self.k_ax.min(), self.k_ax.max(), 'bottom')
        self.set_ticks(self.erg_ax.min(), self.erg_ax.max(), 'left')
        k_min, k_max, e_min, e_max = 0, self.k_ax.size, 0, self.erg_ax.size
        self.cut_panel.setLimits(xMin=k_min, xMax=k_max, yMin=e_min, yMax=e_max, maxXRange=k_max - k_min,
                                 maxYRange=e_max - e_min)

        self.set_mdc_line()

    def set_mdc_line(self):
        if self.erg_ax.min() < 0:
            e_idx = wp.indexof(0, self.erg_ax)
        else:
            e_idx = self.erg_ax.size // 2
        self.mdc_line = InfiniteLine(e_idx, movable=True, angle=0)
        self.mdc_line.setBounds([1, self.erg_ax.size - 1])
        self.mdc_pos = TracedVariable(e_idx, name='pos')
        self.update_allowed_values(1, self.erg_ax.size - 1)
        self.mdc_pos.sig_value_changed.connect(self.update_position)
        self.mdc_line.sigDragged.connect(self.on_dragged)
        self.mdc_line.setPen(BASE_LINECOLOR)
        self.mdc_line.setHoverPen(HOVER_COLOR)

        self.image_e_pos_value_lbl.setText('({:.4f})'.format(self.erg_ax[int(self.mdc_pos.get_value())]))
        self.image_e_pos.setValue(int(self.mdc_pos.get_value()))
        self.cut_panel.addItem(self.mdc_line)

    def set_edc_line(self):
        try:
            self.cut_panel.removeItem(self.edc_line_cut)
            self.mdc_panel.removeItem(self.edc_line_mdc)
        except AttributeError:
            pass

        k_idx = self.k_ax.size // 2
        self.edc_line_cut = InfiniteLine(k_idx, movable=True, angle=90)
        self.edc_line_mdc = InfiniteLine(self.k_ax[k_idx], movable=True, angle=90)
        self.edc_line_cut.setBounds([1, self.k_ax.size - 1])
        self.edc_line_mdc.setBounds([self.k_ax.min(), self.k_ax.max()])
        self.edc_pos = TracedVariable(k_idx, name='pos')
        self.edc_pos.set_allowed_values(np.arange(1, self.k_ax.size - 1, 1))
        # self.edc_pos.sig_value_changed.connect(self.update_position)
        self.edc_line_cut.sigDragged.connect(self.on_dragged_cut_edc)
        self.edc_line_mdc.sigDragged.connect(self.on_dragged_mdc_edc)
        self.edc_line_cut.setPen(BASE_LINECOLOR)
        self.edc_line_cut.setHoverPen(HOVER_COLOR)
        self.edc_line_mdc.setPen((202, 49, 66))
        self.edc_line_mdc.setHoverPen((240, 149, 115))

        self.image_k_pos_value_lbl.setText('({:.4f})'.format(self.k_ax[int(self.edc_pos.get_value())]))
        self.image_k_pos.setValue(int(self.edc_pos.get_value()))
        self.fitting_mu.setValue(self.edc_line_mdc.value())
        self.cut_panel.addItem(self.edc_line_cut)
        self.mdc_panel.addItem(self.edc_line_mdc)

    def set_ticks(self, min_val, max_val, axis):
        """
        Set customized to reflect the dimensions of the physical data
        :param min_val:     float; first axis' value
        :param max_val:     float; last axis' value
        :param axis:        str; axis of which ticks should be put
        """

        plotItem = self.cut_panel.plotItem

        # Remove the old top-axis
        plotItem.layout.removeItem(plotItem.getAxis(axis))
        # Create the new axis and set its range
        new_axis = AxisItem(orientation=axis)
        new_axis.setRange(min_val, max_val)
        plotItem.axes[axis]['item'] = new_axis
        if axis == 'bottom':
            plotItem.layout.addItem(new_axis, *self.cut_panel.main_xaxis_grid)
        else:
            plotItem.layout.addItem(new_axis, *self.cut_panel.main_yaxis_grid)

    def update_position(self):
        self.mdc_line.setValue(self.mdc_pos.get_value())
        self.set_mdc_panel()

    def on_dragged(self):
        self.mdc_pos.set_value(self.mdc_line.value())
        self.image_e_pos_value_lbl.setText('({:.4f})'.format(self.erg_ax[int(self.mdc_pos.get_value())]))
        self.image_e_pos.setValue(int(self.mdc_pos.get_value()))
        # if it's an energy plot, and binning option is active, update also binning boundaries
        if self.image_bin.isChecked():
            pos = self.mdc_line.value()
            n = self.image_bin_n.value()
            self.mdc_up_line.setValue(pos + n)
            self.mdc_down_line.setValue(pos - n)

    def on_dragged_cut_edc(self):
        self.edc_pos.set_value(self.edc_line_cut.value())
        self.edc_line_mdc.setValue(self.k_ax[int(self.edc_line_cut.value())])
        self.image_k_pos_value_lbl.setText('({:.4f})'.format(self.k_ax[int(self.edc_pos.get_value())]))
        self.image_k_pos.setValue(int(self.edc_pos.get_value()))
        self.fitting_mu.setValue(self.edc_line_mdc.value())

    def on_dragged_mdc_edc(self):
        self.edc_pos.set_value(wp.indexof(self.edc_line_mdc.value(), self.k_ax))
        self.edc_line_cut.setValue(wp.indexof(self.edc_line_mdc.value(), self.k_ax))
        self.image_k_pos_value_lbl.setText('({:.4f})'.format(self.k_ax[int(self.edc_pos.get_value())]))
        self.image_k_pos.setValue(int(self.edc_pos.get_value()))
        self.fitting_mu.setValue(self.edc_line_mdc.value())

    def update_mdc_slider(self):
        e = self.image_e_pos.value()
        self.mdc_pos.set_value(e)
        self.set_binning_lines()
        self.image_e_pos_value_lbl.setText('({:.4f})'.format(self.erg_ax[int(self.mdc_pos.get_value())]))

    def update_allowed_values(self, min, max):
        """ Update the allowed values silently.
        This assumes that the displayed image is in pixel coordinates and
        sets the allowed values to the available pixels.
        """
        self.mdc_pos.set_allowed_values(np.arange(min, max + 1, 1))

    def set_binning_lines(self):
        if self.image_bin.isChecked():
            try:
                self.cut_panel.removeItem(self.mdc_up_line)
                self.cut_panel.removeItem(self.mdc_down_line)
            except AttributeError:
                pass

            pos = self.mdc_line.value()
            n = self.image_bin_n.value()
            self.mdc_line.setBounds([n, self.erg_ax.size - n])
            self.update_allowed_values(n, self.erg_ax.size - n)
            self.mdc_up_line = InfiniteLine(pos + n, movable=False, angle=0)
            self.mdc_down_line = InfiniteLine(pos - n, movable=False, angle=0)
            self.mdc_up_line.setPen(color=BINLINES_LINECOLOR, width=1)
            self.mdc_down_line.setPen(color=BINLINES_LINECOLOR, width=1)
            self.cut_panel.addItem(self.mdc_up_line)
            self.cut_panel.addItem(self.mdc_down_line)
        else:
            try:
                self.cut_panel.removeItem(self.mdc_up_line)
                self.cut_panel.removeItem(self.mdc_down_line)
                self.mdc_line.setBounds([1, self.erg_ax.size - 1])
                self.update_allowed_values(1, self.erg_ax.size - 1)
            except AttributeError:
                pass

        self.set_mdc_panel()

    def set_mdc_panel(self):
        mdc_plot = self.mdc_panel
        self.fit_result = None
        self.bgr_fit = None

        for di in mdc_plot.listDataItems():
            try:
                mdc_plot.removeItem(di)
            except AttributeError:
                pass

        e_idx = self.mdc_pos.get_value()
        if self.image_bin.isChecked():
            n = self.image_bin_n.value()
            self.mdc = np.sum(self.data_set.data[0, :, (e_idx - n):(e_idx + n)], axis=1)
        else:
            self.mdc = self.data_set.data[0, :, e_idx]

        range_start, range_stop = self.fitting_range_start.value(), self.fitting_range_stop.value()
        bgr_range_f, bgr_range_s = self.fitting_bgr_range_first.value(), self.fitting_bgr_range_second.value()
        self.fitting_range_start.setRange(self.k_ax.min(), range_stop)
        self.fitting_range_stop.setRange(range_start, self.k_ax.max())
        self.fitting_bgr_range_first.setRange(range_start, range_stop)
        self.fitting_bgr_range_second.setRange(range_start, range_stop)

        mdc_plot.plot(self.k_ax, self.mdc, pen=mkPen('k', width=2))
        shadow_bottom = PlotDataItem([range_start, range_stop], [-self.mdc.max(), -self.mdc.max()])
        shadow_top = PlotDataItem([range_start, range_stop], [2 * self.mdc.max(), 2 * self.mdc.max()])

        try:
            mdc_plot.removeItem(self.shadow)
        except AttributeError:
            pass
        self.shadow = FillBetweenItem(shadow_top, shadow_bottom, mkBrush((229, 229, 229)))

        self.bgr_first = PlotDataItem([bgr_range_f, bgr_range_f],
                                      [-self.mdc.max(), self.mdc[wp.indexof(bgr_range_f, self.k_ax)]],
                                      pen=mkPen('k', style=QtCore.Qt.DashLine))
        self.bgr_second = PlotDataItem([bgr_range_s, bgr_range_s],
                                       [-self.mdc.max(), self.mdc[wp.indexof(bgr_range_s, self.k_ax)]],
                                       pen=mkPen('k', style=QtCore.Qt.DashLine))
        mdc_plot.addItem(self.shadow)
        mdc_plot.addItem(self.bgr_first)
        mdc_plot.addItem(self.bgr_second)
        mdc_plot.setYRange(0, self.mdc.max())

    def fit_bgr(self):
        bgr0, bgr1 = self.fitting_range_start.value(), self.fitting_bgr_range_first.value()
        bgr2, bgr3 = self.fitting_bgr_range_second.value(), self.fitting_range_stop.value()
        bgr0, bgr1 = wp.indexof(bgr0, self.k_ax), wp.indexof(bgr1, self.k_ax)
        bgr2, bgr3 = wp.indexof(bgr2, self.k_ax), wp.indexof(bgr3, self.k_ax)

        bgr_k = np.hstack((self.k_ax[bgr0:bgr1], self.k_ax[bgr2:bgr3]))
        bgr_mdc = np.hstack((self.mdc[bgr0:bgr1], self.mdc[bgr2:bgr3]))
        order = self.fitting_bgr_poly_order.value()
        coefs_bgr = np.polyfit(bgr_k, bgr_mdc, order)

        self.fit_k = self.k_ax[bgr0:bgr3]
        self.bgr_fit = np.poly1d(coefs_bgr)(self.fit_k)

        try:
            self.mdc_panel.removeItem(self.bgr)
        except AttributeError:
            pass

        self.bgr = PlotDataItem(self.fit_k, self.bgr_fit, pen=mkPen('c', width=2))
        self.mdc_panel.addItem(self.bgr)

    def fit_mdc(self):

        try:
            self.mdc_panel.removeItem(self.fit)
        except AttributeError:
            pass

        range0 = wp.indexof(self.fitting_range_start.value(), self.k_ax)
        range1 = wp.indexof(self.fitting_range_stop.value(), self.k_ax)
        k_fit = self.k_ax[range0:range1]
        try:
            mdc_fit = self.mdc[range0:range1] - self.bgr_fit
        except AttributeError:
            mdc_fit = self.mdc[range0:range1]
            self.bgr_fit = np.zeros_like(mdc_fit)
        except TypeError:
            mdc_fit = self.mdc[range0:range1]
            self.bgr_fit = np.zeros_like(mdc_fit)
        except ValueError:
            message = 'Fitting range doesn\'t match background size.'
            self.fitting_message_cell.setText(message)
            return

        a, mu, gamma, alpha, beta = self.fitting_a.value(), self.fitting_mu.value(), self.fitting_gamma.value(), \
                                    self.fitting_alpha.value(), self.fitting_beta.value()
        if alpha == 1:
            fit_alpha = False
        else:
            fit_alpha = True
        if beta == 1:
            fit_beta = False
        else:
            fit_beta = True
        self.set_fit_fun(fit_alpha, fit_beta)

        try:
            self.p, cov = curve_fit(self.fit_fun, k_fit, mdc_fit, p0=self.p0)
            message = self.set_fitting_message(fit_alpha, fit_beta)
            self.fitting_message_cell.setText(message)
        except RuntimeWarning:
            message = 'Runtime: Couldn\'t fit with current parameters.'
            self.fitting_message_cell.setText(message)
            return
        except OptimizeWarning:
            message = 'Optimization: Couldn\'t fit with current parameters.'
            self.fitting_message_cell.setText(message)
            return

        self.prepare_fitting_results(fit_alpha, fit_beta)
        res_func = lambda x: self.fit_fun(x, *self.p)
        fit = res_func(k_fit) + self.bgr_fit

        self.fit = PlotDataItem(k_fit, fit, pen=mkPen('m', width=2))
        self.mdc_panel.addItem(self.fit)

    def set_fit_fun(self, fit_alpha, fit_beta):
        self.p0 = [self.fitting_a.value(), self.fitting_mu.value(), self.fitting_gamma.value()]
        if fit_alpha and fit_beta:
            self.fit_fun = lambda x, a0, mu, gamma, alpha, beta: wp.asym_lorentzian(x, a0, mu, gamma, alpha, beta)
            self.p0.append(self.fitting_alpha.value())
            self.p0.append(self.fitting_beta.value())
        elif fit_alpha and not fit_beta:
            self.fit_fun = lambda x, a0, mu, gamma, alpha: wp.asym_lorentzian(x, a0, mu, gamma, alpha=alpha)
            self.p0.append(self.fitting_alpha.value())
        elif not fit_alpha and fit_beta:
            self.fit_fun = lambda x, a0, mu, gamma, beta: wp.asym_lorentzian(x, a0, mu, gamma, beta=beta)
            self.p0.append(self.fitting_beta.value())
        else:
            self.fit_fun = lambda x, a0, mu, gamma: wp.asym_lorentzian(x, a0, mu, gamma)

    def set_fitting_message(self, fit_alpha, fit_beta):
        e = self.erg_ax[int(self.mdc_pos.get_value())]
        message = 'E = {:.4f};  a = {:.1f};  \u03BC = {:.4f};  \u0393 = {:.4f}'.format(e, self.p[0], self.p[1], self.p[2])
        if fit_alpha:
            message += ';  \u03B1 = {:.4f}'.format(self.p[3])
        else:
            message += ';  \u03B1 = -'

        if fit_beta:
            message += ';  \u03B2 = {:.4f}'.format(self.p[-1])
        else:
            message += ';  \u03B2 = -'

        return message

    def prepare_fitting_results(self, fit_alpha, fit_beta):
        e = self.erg_ax[int(self.mdc_pos.get_value())]
        res = [e, self.p[0], self.p[1], self.p[2]]
        if fit_alpha:
            res.append(self.p[3])
        else:
            res.append(1)

        if fit_beta:
            res.append(self.p[-1])
        else:
            res.append(1)
        self.fit_result = np.array(res)

    def append_fit_results(self):

        if self.fit_result is None:
            no_fit_box = QMessageBox()
            no_fit_box.setIcon(QMessageBox.Information)
            no_fit_box.setText('No result to save.')
            no_fit_box.setStandardButtons(QMessageBox.Ok)
            if no_fit_box.exec() == QMessageBox.Ok:
                return

        if self.fit_results is None:
            self.fit_results = self.fit_result
        else:
            self.fit_results = np.vstack((self.fit_results, self.fit_result))

        if len(self.fit_results.shape) == 1:
            pass
        else:
            self.fit_results = np.flip(np.sort(self.fit_results, axis=0), axis=0)

        try:
            self.cut_panel.removeItem(self.fit_points)
        except AttributeError:
            pass

        try:
            fit_points_k = np.array([wp.indexof(ki, self.k_ax) for ki in self.fit_results[:, 2]])
            fit_points_e = np.array([wp.indexof(ei, self.erg_ax) for ei in self.fit_results[:, 0]])
            self.fit_points = ScatterPlotItem(fit_points_k, fit_points_e, symbol='h')
            self.cut_panel.addItem(self.fit_points)
        except IndexError:
            pass

    def save_fit_results(self):

        fname = self.title[:-13] + '_mdc_fit_results.txt'
        if os.path.exists(fname):
            no_fit_box = QMessageBox()
            no_fit_box.setIcon(QMessageBox.Information)
            no_fit_box.setText('File already exists, want to overwrite?')
            no_fit_box.setStandardButtons(QMessageBox.Ok | QMessageBox.Cancel)
            if no_fit_box.exec() == QMessageBox.Cancel:
                return
            else:
                pass

        f = open(fname, 'w+')
        line = 'E [eV]\t\t\ta [a.u.] \t\t mu [1/A]\t\tgamma [1/A]\t\talpha\tbeta [a.u]\n'
        f.write(line)
        try:
            for result in self.fit_results:
                line = ''
                for entry in result:
                    line += str(entry) + '\t'
                f.write(line + '\n')
        except TypeError:
            for entry in self.fit_results:
                line += str(entry) + '\t'
            f.write(line + '\n')
        f.close()

    def edit_fit_results(self):

        fname = self.title[:-13] + '_mdc_fit_results.txt'
        if not os.path.exists(fname):
            no_file_box = QMessageBox()
            no_file_box.setIcon(QMessageBox.Information)
            no_file_box.setText('File not found.')
            no_file_box.setStandardButtons(QMessageBox.Ok)
            if no_file_box.exec() == QMessageBox.Ok:
                return

        os.system('open ' + fname)

    def load_update_fit_result(self):

        fname = self.title[:-13] + '_mdc_fit_results.txt'
        if not os.path.exists(fname):
            no_file_box = QMessageBox()
            no_file_box.setIcon(QMessageBox.Information)
            no_file_box.setText('File not found.')
            no_file_box.setStandardButtons(QMessageBox.Ok)
            if no_file_box.exec() == QMessageBox.Ok:
                return

        overwriting_box = QMessageBox()
        overwriting_box.setIcon(QMessageBox.Question)
        overwriting_box.setText('Fit results will be overwritten, sure to proceed?')
        overwriting_box.setStandardButtons(QMessageBox.Ok | QMessageBox.Cancel)
        if overwriting_box.exec() == QMessageBox.Cancel:
            return
        else:
            pass

        f = open(fname)
        lines = f.readlines()
        self.fit_results = None
        for line in lines[1:]:
            row = np.array([float(entry) for entry in line.split('\t')[:-1]])
            if self.fit_results is None:
                self.fit_results = row
            else:
                self.fit_results = np.vstack((self.fit_results, row))

        f.close()

        try:
            self.cut_panel.removeItem(self.fit_points)
        except AttributeError:
            pass

        try:
            fit_points_k = np.array([wp.indexof(ki, self.k_ax) for ki in self.fit_results[:, 2]])
            fit_points_e = np.array([wp.indexof(ei, self.erg_ax) for ei in self.fit_results[:, 0]])
            self.fit_points = ScatterPlotItem(fit_points_k, fit_points_e, symbol='h')
            self.cut_panel.addItem(self.fit_points)
        except IndexError:
            pass

    def close(self):
        self.destroy()
        self.data_viewer.thread[self.thread_index].quit()
        self.data_viewer.thread[self.thread_index].wait()
        del(self.data_viewer.thread[self.thread_index])
        del(self.data_viewer.data_viewers[self.thread_index])





# main for calling files' openers

import pkg_resources
import os
import sys

from PyQt5.QtWidgets import QApplication

import piva._2Dviewer as p2d
import piva._3Dviewer as p3d
import piva.data_loader as dl
from piva.data_browser import DataBrowser

# to fix bugs in Big Siur
os.environ['QT_MAC_WANTS_LAYER'] = '1'

# TODO DOCUMANTATION !!!!!!!!!!!
# TODO glue scans together


def db():
    version = pkg_resources.require('piva')[0].version
    print(f'+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++')
    print(f'        PIVA - Python Interactive Viewer for Arpes         ')
    print(f'                    Version {version}                      ')
    print(f'+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++')
    app = QApplication(sys.argv)
    window = DataBrowser()
    sys.exit(app.exec_())


def pickle_h5():
    fname = sys.argv[1]
    data = dl.load_data(fname)
    dl.dump(data, fname[:-3])


if __name__ == "__main__" :
    db()



import argparse
import datetime
from copy import deepcopy

import numpy as np
import pyqtgraph as pg
from pyqtgraph.Qt import QtGui, QtWidgets
from PyQt5.QtGui import QFont, QColor
from PyQt5.QtWidgets import QColorDialog, QFileDialog, QWidget, \
    QDoubleSpinBox, QLineEdit, QPushButton, QLabel, QComboBox, QSpinBox, \
    QCheckBox, QMessageBox

import piva.arpys_wp as wp
import piva.data_loader as dl
import piva._2Dviewer as p2d
import piva._3Dviewer as p3d
from piva.imageplot import bold_font

MDC_PANEL_BGR = (236, 236, 236)


class PlotTool(QtWidgets.QMainWindow):

    def __init__(self, data_browser, title=None):
        super(PlotTool, self).__init__()

        self.central_widget = QWidget()
        self.plotting_tool_layout = QtWidgets.QGridLayout()
        self.central_widget.setLayout(self.plotting_tool_layout)
        self.tabs = QtWidgets.QTabWidget()

        self.main_utils = QWidget()
        self.plot_panel = pg.PlotWidget(background=MDC_PANEL_BGR)

        self.data_browser = data_browser
        self.title = title
        self.data_items = {}

        self.plot_panel_design = {}
        self.init_panel_design()

        self.marker_1 = {}
        self.marker_2 = {}
        self.init_markers()

        self.annotations = {}

        self.align()
        self.set_main_utils()
        self.set_tabs()
        self.set_ds_dv_list()

        self.initUI()
        self.setCentralWidget(self.central_widget)
        self.setWindowTitle(self.title)
        self.show()

    def initUI(self):

        self.main_added.currentIndexChanged.connect(self.current_selection_changed)
        self.main_save_button.clicked.connect(self.save)
        self.main_load_button.clicked.connect(self.load)
        self.main_close_button.clicked.connect(self.close)

        self.ds_update_lists.clicked.connect(self.set_ds_dv_list)
        self.ds_dv.currentIndexChanged.connect(self.set_ds_dv_plot_list)
        self.ds_add_button.clicked.connect(self.add_dataset)
        self.ds_remove_button.clicked.connect(self.remove_dataset)

        self.ec_color.clicked.connect(self.set_color)
        self.ec_width.valueChanged.connect(lambda: self.set_pen(loading=False))
        self.ec_style.currentIndexChanged.connect(lambda: self.set_pen(loading=False))
        self.ec_offset_x.valueChanged.connect(lambda: self.update_dataset(loading=False))
        self.ec_offset_y.valueChanged.connect(lambda: self.update_dataset(loading=False))
        self.ec_normalize.stateChanged.connect(lambda: self.update_dataset(loading=False))
        self.ec_scale.valueChanged.connect(lambda: self.update_dataset(loading=False))
        self.ec_reset.clicked.connect(self.reset_scaling)

        self.ep_bgr_color.clicked.connect(lambda: self.set_bgr_color(loading=False))
        self.ep_axes_color.clicked.connect(lambda: self.set_axes_color(loading=False))
        self.ep_ticks_size.valueChanged.connect(lambda: self.set_plot_layout(loading=False))
        self.ep_xlabel.textChanged.connect(lambda: self.set_plot_layout(loading=False))
        self.ep_ylabel.textChanged.connect(lambda: self.set_plot_layout(loading=False))
        self.ep_labels_font_size.valueChanged.connect(lambda: self.set_plot_layout(loading=False))

        self.marker_1['button'].clicked.connect(lambda: self.drop_marker(marker=self.marker_1))
        self.marker_2['button'].clicked.connect(lambda: self.drop_marker(marker=self.marker_2))
        self.marker_1['x'].valueChanged.connect(lambda: self.update_marker(marker=self.marker_1))
        self.marker_2['x'].valueChanged.connect(lambda: self.update_marker(marker=self.marker_2))

        self.ann_add_update.clicked.connect(self.add_update_annotation)
        self.ann_remove.clicked.connect(self.remove_annotation)
        self.ann_color.clicked.connect(self.set_annotation_color)
        self.ann_added.currentIndexChanged.connect(self.ann_selection_changed)

    def init_panel_design(self):
        self.plot_panel_design['bgr_color'] = QColor()
        self.plot_panel_design['bgr_color'].setRgb(MDC_PANEL_BGR[0], MDC_PANEL_BGR[1], MDC_PANEL_BGR[2])
        self.plot_panel_design['axes_color'] = QColor()
        self.plot_panel_design['axes_color'].setRgb(150, 150, 150)
        self.plot_panel_design['ticks_fsize'] = QFont()
        self.plot_panel_design['ticks_fsize'].setPointSize(14)
        self.plot_panel_design['labels_fsize'] = 14

    def init_markers(self):
        max_w = 80

        self.marker_1['marker'] = pg.ScatterPlotItem()
        self.marker_1['dropped'] = False
        self.marker_1['idx'] = 1
        self.marker_1['x'] = QDoubleSpinBox()
        self.set_qspinbox(self.marker_1['x'], [-1e6, 1e6], value=0, decimals=6, step=0.01)
        self.marker_1['y'] = QLineEdit()
        self.marker_1['y'].setReadOnly(True)
        self.marker_1['y'].setText('-')
        self.marker_1['y'].setMaximumWidth(max_w)
        self.marker_1['button'] = QPushButton('drop')
        self.marker_1['dumped_at'] = None

        self.marker_2['marker'] = pg.ScatterPlotItem()
        self.marker_2['dropped'] = False
        self.marker_2['idx'] = 2
        self.marker_2['x'] = QDoubleSpinBox()
        self.set_qspinbox(self.marker_2['x'], [-1e6, 1e6], value=0, decimals=6, step=0.01)
        self.marker_2['y'] = QLineEdit()
        self.marker_2['y'].setReadOnly(True)
        self.marker_2['y'].setText('-')
        self.marker_2['y'].setMaximumWidth(max_w)
        self.marker_2['button'] = QPushButton('drop')
        self.marker_2['dumped_at'] = None

    def align(self):
        ptl = self.plotting_tool_layout

        ptl.addWidget(self.main_utils,  0, 0, 4, 1)
        ptl.addWidget(self.tabs,        0, 1, 4, 7)
        ptl.addWidget(self.plot_panel,  4, 0, 8, 8)

    def set_tabs(self):
        self.set_datasets_tab()
        self.set_edit_curves_tab()
        self.set_edit_plot_tab()
        self.set_markers_tab()
        self.set_annotate_tab()

    def set_main_utils(self):
        # self.main_utils = QWidget()
        mul = QtWidgets.QVBoxLayout()

        self.main_added_lbl = QLabel('Curves:')
        self.main_added = QComboBox()

        self.main_save_button = QPushButton('save')
        self.main_load_button = QPushButton('load')
        self.main_close_button = QPushButton('close')

        mul.addWidget(self.main_added_lbl)
        mul.addWidget(self.main_added)
        mul.addWidget(self.main_save_button)
        mul.addWidget(self.main_load_button)
        mul.addWidget(self.main_close_button)

        self.main_utils.layout = mul
        self.main_utils.setLayout(mul)

    def set_datasets_tab(self):
        # create elements
        self.datasets_tab = QWidget()
        dtl = QtWidgets.QGridLayout()

        self.ds_dv_label = QLabel('dataset:')
        self.ds_dv = QComboBox()
        self.ds_dv.addItem('custom')
        self.ds_dv_plot_label = QLabel('curve:')
        self.ds_dv_plot = QComboBox()

        self.ds_update_lists = QPushButton('update')
        self.ds_add_button = QPushButton('add')
        self.ds_remove_button = QPushButton('remove current')

        self.ds_custom_x_lbl = QLabel('x:')
        self.ds_custom_x = QLineEdit()
        self.ds_custom_y_lbl = QLabel('y:')
        self.ds_custom_y = QLineEdit()
        self.ds_custom_name_lbl = QLabel('name:')
        self.ds_custom_name = QLineEdit()

        row = 0
        dtl.addWidget(self.ds_dv_label,             row, 0)
        dtl.addWidget(self.ds_dv,                   row, 1)
        dtl.addWidget(self.ds_dv_plot_label,        row, 2)
        dtl.addWidget(self.ds_dv_plot,              row, 3)
        dtl.addWidget(self.ds_add_button,           row, 4)
        dtl.addWidget(self.ds_update_lists,         row, 5)

        row = 1
        dtl.addWidget(self.ds_custom_name_lbl,      row, 0)
        dtl.addWidget(self.ds_custom_name,          row, 1)
        dtl.addWidget(self.ds_remove_button,        row, 4, 1, 2)

        row = 2
        dtl.addWidget(self.ds_custom_x_lbl,         row, 0)
        dtl.addWidget(self.ds_custom_x,             row, 1)
        dtl.addWidget(self.ds_custom_y_lbl,         row, 2)
        dtl.addWidget(self.ds_custom_y,             row, 3)

        # dummy_lbl = QLabel('')
        # dtl.addWidget(dummy_lbl, 2, 0, 1, 6)

        self.datasets_tab.layout = dtl
        self.datasets_tab.setLayout(dtl)
        self.tabs.addTab(self.datasets_tab, 'Add/Remove data')

    def set_edit_curves_tab(self):
        # create elements
        self.edit_curves_tab = QWidget()
        etl = QtWidgets.QGridLayout()

        self.ec_color_lbl = QLabel("color:")
        self.ec_color = QPushButton('')
        self.ec_color.setStyleSheet("background-color: black")
        self.ec_color.setFixedSize(25, 15)
        self.ec_width_lbl = QLabel('width')
        self.ec_width = QSpinBox()
        self.set_qspinbox(self.ec_width, box_range=[1, 50], value=3)
        self.ec_style_lbl = QLabel('line style:')
        self.ec_style = QComboBox()
        line_types = ['-', '---', '...', '-.-', '-..']
        for lt in line_types:
            self.ec_style.addItem(lt)

        self.ec_offset_x_lbl = QLabel('x offset:')
        self.ec_offset_x = QDoubleSpinBox()
        self.set_qspinbox(self.ec_offset_x, box_range=[-1e6, 1e6], value=0, decimals=5, step=1e-3)
        self.ec_offset_y_lbl = QLabel('y offset:')
        self.ec_offset_y = QDoubleSpinBox()
        self.set_qspinbox(self.ec_offset_y, box_range=[-1e6, 1e6], value=0, decimals=4, step=0.1)

        self.ec_normalize = QCheckBox('normalize')
        self.ec_scale_lbl = QLabel('scale:')
        self.ec_scale = QDoubleSpinBox()
        self.set_qspinbox(self.ec_scale, box_range=[0, 1e6], value=1,  decimals=1, step=1)
        self.ec_reset = QPushButton('reset')

        row = 0
        etl.addWidget(self.ec_color_lbl,              row, 0)
        etl.addWidget(self.ec_color,                  row, 1)
        etl.addWidget(self.ec_width_lbl,              row, 2)
        etl.addWidget(self.ec_width,                  row, 3)
        etl.addWidget(self.ec_style_lbl,              row, 4)
        etl.addWidget(self.ec_style,                  row, 5)

        row = 1
        etl.addWidget(self.ec_offset_x_lbl,           row, 0)
        etl.addWidget(self.ec_offset_x,               row, 1, 1, 2)
        etl.addWidget(self.ec_offset_y_lbl,           row, 3)
        etl.addWidget(self.ec_offset_y,               row, 4, 1, 2)

        row = 2
        etl.addWidget(self.ec_scale_lbl,              row, 0)
        etl.addWidget(self.ec_scale,                  row, 1)
        etl.addWidget(self.ec_normalize,              row, 2)
        etl.addWidget(self.ec_reset,                  row, 3)

        # dummy_lbl = QLabel('')
        # etl.addWidget(dummy_lbl, 2, 0, 1, 6)

        self.edit_curves_tab.layout = etl
        self.edit_curves_tab.setLayout(etl)
        self.tabs.addTab(self.edit_curves_tab, 'Edit curves')

    def set_edit_plot_tab(self):
        # create elements
        self.edit_plot_tab = QWidget()
        eptl = QtWidgets.QGridLayout()

        self.ep_bgr_color_lbl = QLabel("bgr color:")
        self.ep_bgr_color = QPushButton('')
        self.ep_bgr_color.setStyleSheet(f'background-color: rgb{MDC_PANEL_BGR}')
        self.ep_bgr_color.setFixedSize(25, 15)
        self.ep_axes_color_lbl = QLabel("axes color:")
        self.ep_axes_color = QPushButton('')
        self.ep_axes_color.setStyleSheet(f'background-color: rgb(150, 150, 150)')
        self.ep_axes_color.setFixedSize(25, 15)
        self.ep_ticks_size_lbl = QLabel('ticks font size:')
        self.ep_ticks_size = QSpinBox()
        self.set_qspinbox(self.ep_ticks_size, box_range=[1, 100], value=14)

        self.ep_labels_font_size_lbl = QLabel('labels font size:')
        self.ep_labels_font_size = QSpinBox()
        self.set_qspinbox(self.ep_labels_font_size, box_range=[1, 100], value=14)

        self.ep_xlabel_lbl = QLabel('x label:')
        self.ep_xlabel = QLineEdit()
        self.ep_ylabel_lbl = QLabel('y label:')
        self.ep_ylabel = QLineEdit()

        row = 0
        eptl.addWidget(self.ep_bgr_color_lbl,               row, 0)
        eptl.addWidget(self.ep_bgr_color,                   row, 1)
        eptl.addWidget(self.ep_axes_color_lbl,              row, 2)
        eptl.addWidget(self.ep_axes_color,                  row, 3)
        eptl.addWidget(self.ep_ticks_size_lbl,              row, 4)
        eptl.addWidget(self.ep_ticks_size,                  row, 5)

        row = 1
        eptl.addWidget(self.ep_xlabel_lbl,                  row, 0)
        eptl.addWidget(self.ep_xlabel,                      row, 1, 1, 2)
        eptl.addWidget(self.ep_ylabel_lbl,                  row, 3)
        eptl.addWidget(self.ep_ylabel,                      row, 4, 1, 2)

        row = 2
        eptl.addWidget(self.ep_labels_font_size_lbl,        row, 4)
        eptl.addWidget(self.ep_labels_font_size,            row, 5)

        self.edit_plot_tab.layout = eptl
        self.edit_plot_tab.setLayout(eptl)
        self.tabs.addTab(self.edit_plot_tab, 'Edit plot')

    def set_markers_tab(self):
        # create elements
        self.markers_tab = QWidget()
        mtl = QtWidgets.QGridLayout()
        max_w = 80

        self.markers_x_lbl = QLabel("x:")
        self.markers_y_lbl = QLabel("y:")
        self.markers_dx_lbl = QLabel("dx:")
        self.markers_dy_lbl = QLabel("dy:")

        self.marker_1_lbl = QLabel('1:')
        self.marker_1_lbl.setFont(bold_font)
        self.markers_dx = QLineEdit()
        self.markers_dx.setReadOnly(True)
        self.markers_dx.setMaximumWidth(max_w)
        self.markers_dy = QLineEdit()
        self.markers_dy.setReadOnly(True)
        self.markers_dy.setMaximumWidth(max_w)

        self.marker_2_lbl = QLabel('2:')
        self.marker_2_lbl.setFont(bold_font)

        row = 0
        mtl.addWidget(self.markers_x_lbl,       row, 1)
        mtl.addWidget(self.markers_y_lbl,       row, 2)
        mtl.addWidget(self.markers_dx_lbl,      row, 3)
        mtl.addWidget(self.markers_dy_lbl,      row, 4)

        row = 1
        mtl.addWidget(self.marker_1_lbl,        row, 0)
        mtl.addWidget(self.marker_1['x'],       row, 1)
        mtl.addWidget(self.marker_1['y'],       row, 2)
        mtl.addWidget(self.markers_dx,          row, 3)
        mtl.addWidget(self.markers_dy,          row, 4)
        mtl.addWidget(self.marker_1['button'],  row, 5)

        row = 2
        mtl.addWidget(self.marker_2_lbl,        row, 0)
        mtl.addWidget(self.marker_2['x'],       row, 1)
        mtl.addWidget(self.marker_2['y'],       row, 2)
        mtl.addWidget(self.marker_2['button'],  row, 5)

        self.markers_tab.layout = mtl
        self.markers_tab.setLayout(mtl)
        self.tabs.addTab(self.markers_tab, 'Markers')

    def set_annotate_tab(self):
        # create elements
        self.ann_tab = QWidget()
        atl = QtWidgets.QGridLayout()

        self.ann_name_lbl = QLabel('name:')
        self.ann_name = QLineEdit()
        self.ann_text_lbl = QLabel('text:')
        self.ann_text = QLineEdit()

        self.ann_added_lbl = QLabel('added:')
        self.ann_added = QComboBox()
        self.ann_fsize_lbl = QLabel('font size:')
        self.ann_fsize = QSpinBox()
        self.set_qspinbox(self.ann_fsize, box_range=[2, 50], value=14)
        self.ann_color_lbl = QLabel('color:')
        self.ann_color = QPushButton('')
        self.ann_color.setStyleSheet("background-color: black")
        self.ann_color.setFixedSize(25, 15)

        self.ann_x_lbl = QLabel('x:')
        self.ann_x = QDoubleSpinBox()
        self.set_qspinbox(self.ann_x, box_range=[-1e6, 1e6], value=-0.5, step=0.01)
        self.ann_y_lbl = QLabel('y:')
        self.ann_y = QDoubleSpinBox()
        self.set_qspinbox(self.ann_y, box_range=[-1e6, 1e6], value=-0.5, step=0.01)
        self.ann_add_update = QPushButton('add/update')
        self.ann_remove = QPushButton('delete')

        row = 0
        atl.addWidget(self.ann_name_lbl,        row, 0)
        atl.addWidget(self.ann_name,            row, 1, 1, 2)
        atl.addWidget(self.ann_text_lbl,        row, 3)
        atl.addWidget(self.ann_text,            row, 4, 1, 2)

        row = 1
        atl.addWidget(self.ann_added_lbl,       row, 0)
        atl.addWidget(self.ann_added,           row, 1)
        atl.addWidget(self.ann_fsize_lbl,       row, 2)
        atl.addWidget(self.ann_fsize,           row, 3)
        atl.addWidget(self.ann_color_lbl,       row, 4)
        atl.addWidget(self.ann_color,           row, 5)

        row = 2
        atl.addWidget(self.ann_x_lbl,           row, 0)
        atl.addWidget(self.ann_x,               row, 1)
        atl.addWidget(self.ann_y_lbl,           row, 2)
        atl.addWidget(self.ann_y,               row, 3)
        atl.addWidget(self.ann_add_update,      row, 4)
        atl.addWidget(self.ann_remove,          row, 5)

        self.ann_tab.layout = atl
        self.ann_tab.setLayout(atl)
        self.tabs.addTab(self.ann_tab, 'Annotate')

    def set_ds_dv_list(self):
        # clear old lists
        self.dv_list = []
        try:
            self.ds_dv.clear()
            self.ds_dv_plot.clear()
        except AttributeError:
            pass

        self.ds_dv.addItem('custom')
        self.dv_list.append('custom')
        dv = self.data_browser.data_viewers
        for dvi in dv.keys():
            self.dv_list.append(dvi)
            dvi_lbl = dvi.split('/')[-1]
            self.ds_dv.addItem(dvi_lbl)
            if isinstance(self.data_browser.data_viewers[dvi], MainWindow3D):
                for key in self.data_browser.data_viewers[dvi].data_viewers.keys():
                    key_lbl = key.split('/')[-1]
                    self.dv_list.append(key)
                    self.ds_dv.addItem(key_lbl)

    def set_ds_dv_plot_list(self):

        self.ds_dv_plot.clear()
        try:
            idx = self.ds_dv.currentIndex()
            dv_lbl = self.dv_list[idx]
        except IndexError:
            return

        dv = self.data_browser.data_viewers[dv_lbl]

        if isinstance(dv, p2d.MainWindow2D):
            self.ds_dv_plot.addItem('edc')
            self.ds_dv_plot.addItem('mdc')
            for key in dv.data_viewers.keys():
                if 'edc_viewer' in key:
                    self.ds_dv_plot.addItem('edc_fitter')
                if 'mdc_viewer' in key:
                    self.ds_dv_plot.addItem('mdc_fitter')
        elif isinstance(dv, MainWindow3D):
            self.ds_dv_plot.addItem('main edc')
            self.ds_dv_plot.addItem('single point edc')
            self.ds_dv_plot.addItem('vertical (analyzer)')
            self.ds_dv_plot.addItem('horizontal (scanned)')

    def add_dataset(self):

        idx = self.ds_dv.currentIndex()

        try:
            dv_lbl = self.dv_list[idx]
        except AttributeError:
            no_entries_box = QMessageBox()
            no_entries_box.setIcon(QMessageBox.Information)
            no_entries_box.setWindowTitle('Doh.')
            no_entries_box.setText('No entries to add.  Make sure to update.')
            no_entries_box.setStandardButtons(QMessageBox.Ok)
            if no_entries_box.exec() == QMessageBox.Ok:
                return
        except IndexError:
            no_entries_box = QMessageBox()
            no_entries_box.setIcon(QMessageBox.Information)
            no_entries_box.setWindowTitle('Doh.')
            no_entries_box.setText('Something went wrong.  Open new data set and update.')
            no_entries_box.setStandardButtons(QMessageBox.Ok)
            if no_entries_box.exec() == QMessageBox.Ok:
                return

        if dv_lbl == 'custom':
            plot = self.ds_custom_name.text()
            if plot == '':
                no_name_box = QMessageBox()
                no_name_box.setIcon(QMessageBox.Information)
                no_name_box.setWindowTitle('Doh.')
                no_name_box.setText('Need to specify data set\'s name.')
                no_name_box.setStandardButtons(QMessageBox.Ok)
                if no_name_box.exec() == QMessageBox.Ok:
                    return
        else:
            try:
                dv = self.data_browser.data_viewers[dv_lbl]
                plot = self.ds_dv_plot.currentText()
            except KeyError:
                no_dv_box = QMessageBox()
                no_dv_box.setIcon(QMessageBox.Information)
                no_dv_box.setWindowTitle('Doh.')
                no_dv_box.setText('It seems the data set has been closed.  Make sure to update.')
                no_dv_box.setStandardButtons(QMessageBox.Ok)
                if no_dv_box.exec() == QMessageBox.Ok:
                    return

        if dv_lbl == 'custom':
            x = self.ds_custom_x.text().split(' ')
            y = self.ds_custom_y.text().split(' ')
            x = np.array([float(xi) for xi in x])
            y = np.array([float(yi) for yi in y])
            data_item_lbl = plot
        else:
            if isinstance(dv, p2d.MainWindow2D):
                x, y = self.get_data_from_2dviewer(dv_lbl, plot)
            elif isinstance(dv, p3d.MainWindow3D):
                x, y = self.get_data_from_3dviewer(dv_lbl, plot)
            data_item_lbl = dv_lbl.split('/')[-1] + ' - ' + plot

        if data_item_lbl in self.data_items.keys():
            data_item_colision_box = QMessageBox()
            data_item_colision_box.setIcon(QMessageBox.Question)
            data_item_colision_box.setWindowTitle('Ooh.')
            data_item_colision_box.setText('Curve with this name already has been added.  Want to add another one?')
            data_item_colision_box.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
            if data_item_colision_box.exec() == QMessageBox.Yes:
                lbl, lbl_return_val = QtWidgets.QInputDialog.getText(self, '', 'Then add specific label:', QLineEdit.Normal, '1')
                data_item_lbl = data_item_lbl + ' (' + lbl + ')'
                if not lbl_return_val:
                    return
            else:
                return

        self.data_items[data_item_lbl] = {}
        self.data_items[data_item_lbl]['data_item'] = CustomDataItem(x, y, pen=pg.mkPen('k', width=2))
        self.data_items[data_item_lbl]['org_data'] = [deepcopy(x), deepcopy(y)]
        self.data_items[data_item_lbl]['color'] = 'k'
        self.data_items[data_item_lbl]['width'] = 2
        self.data_items[data_item_lbl]['line_style'] = '-'
        self.data_items[data_item_lbl]['x_off'] = 0
        self.data_items[data_item_lbl]['y_off'] = 0
        self.data_items[data_item_lbl]['scale'] = 1
        self.data_items[data_item_lbl]['normalize'] = False

        self.plot_panel.addItem(self.data_items[data_item_lbl]['data_item'])
        self.main_added.addItem(data_item_lbl)

    def remove_dataset(self):

        item = self.main_added.currentText()
        idx = self.main_added.currentIndex()
        self.plot_panel.removeItem(self.data_items[item]['data_item'])
        self.main_added.removeItem(idx)
        del(self.data_items[item])

    def current_selection_changed(self):
        try:
            di = self.data_items[self.main_added.currentText()]
        except KeyError:
            return

        self.change_without_emitting_signal(self.ec_width, di['width'])
        self.change_without_emitting_signal(self.ec_style, di['line_style'])
        self.change_without_emitting_signal(self.ec_offset_x, di['x_off'])
        self.change_without_emitting_signal(self.ec_offset_y, di['y_off'])
        self.change_without_emitting_signal(self.ec_scale, di['scale'])
        self.change_without_emitting_signal(self.ec_normalize, di['normalize'])

        clr = di['color']
        try:
            clr = di['color'].getRgb()[:3]
        except AttributeError:
            pass
        self.ec_color.setStyleSheet(f'background-color: rgb{clr}')

    def get_data_from_2dviewer(self, dv_lbl, plot):
        dv = self.data_browser.data_viewers[dv_lbl]

        if plot == 'edc':
            if dv.new_energy_axis is None:
                return dv.data_handler.axes[0], dv.edc
            else:
                return dv.new_energy_axis, dv.edc
        elif plot == 'mdc':
            if dv.k_axis is None:
                return dv.data_handler.axes[1], dv.mdc
            else:
                return dv.k_axis, dv.mdc
        elif plot == 'edc_fitter':
            edc_fitter = dv.data_viewers[dv_lbl + '_edc_viewer']
            return edc_fitter.edc_erg_ax, edc_fitter.edc
        elif plot == 'mdc_fitter':
            mdc_fitter = dv.data_viewers[dv_lbl + '_mdc_viewer']
            return mdc_fitter.k_ax, mdc_fitter.mdc

    def get_data_from_3dviewer(self, dv_lbl, plot):
        dv = self.data_browser.data_viewers[dv_lbl]

        if plot == 'main edc':
            edc = dv.plot_z.get_data()[1]
            if dv.new_energy_axis is None:
                return dv.data_set.zscale, edc
            else:
                return dv.new_energy_axis, edc
        elif plot == 'single point edc':
            edc = dv.sp_EDC.getData()[1]
            if dv.new_energy_axis is None:
                return dv.data_set.zscale, edc
            else:
                return dv.new_energy_axis, edc
        elif plot == 'vertical (analyzer)':
            return dv.data_set.yscale, dv.plot_y_data
        elif plot == 'horizontal (scanned)':
            return dv.data_set.xscale, dv.plot_x_data

    def set_color(self):
        cd = QColorDialog.getColor()
        cd_str = 'rgb' + str(cd.getRgb()[:3])
        self.ec_color.setStyleSheet(f'background-color: {cd_str}')
        data_item_lbl = self.main_added.currentText()
        try:
            self.data_items[data_item_lbl]['color'] = cd
        except KeyError:
            return

        self.set_pen()

    def set_pen(self, loading=False):
        data_item_lbl = self.main_added.currentText()
        if loading:
            w = self.data_items[data_item_lbl]['width']
            self.ec_style.setCurrentText(self.data_items[data_item_lbl]['line_style'])
            style, style_lbl = self.get_line_style()
            cd = self.data_items[data_item_lbl]['color']
        else:
            w = self.ec_width.value()
            style, style_lbl = self.get_line_style()
            cd = self.data_items[data_item_lbl]['color']

        self.data_items[data_item_lbl]['data_item'].setPen(color=cd, width=w, style=style)
        # self.data_items[data_item_lbl]['color'] = cd
        self.data_items[data_item_lbl]['width'] = w
        self.data_items[data_item_lbl]['line_style'] = style_lbl

    def get_line_style(self):
        s = self.ec_style.currentText()
        if s == '-':
            style = QtCore.Qt.SolidLine
        elif s == '---':
            style = QtCore.Qt.DashLine
        elif s == '...':
            style = QtCore.Qt.DotLine
        elif s == '-.-':
            style = QtCore.Qt.DashDotLine
        elif s == '-..':
            style = QtCore.Qt.DashDotDotLine

        return style, s

    def update_dataset(self, loading=False):
        data_item_lbl = self.main_added.currentText()
        if loading:
            x_off = self.data_items[data_item_lbl]['x_off']
            y_off = self.data_items[data_item_lbl]['y_off']
            scale = self.data_items[data_item_lbl]['scale']
            self.ec_normalize.setChecked(self.data_items[data_item_lbl]['normalize'])
        else:
            x_off = self.ec_offset_x.value()
            y_off = self.ec_offset_y.value()
            scale = self.ec_scale.value()

        org = deepcopy(self.data_items[data_item_lbl]['org_data'])
        x, y = org[0], org[1]
        if self.ec_normalize.isChecked():
            y = wp.normalize(y)
            self.data_items[data_item_lbl]['normalize'] = True
        y *= scale
        x += x_off
        y += y_off
        self.data_items[data_item_lbl]['data_item'].setData(x, y)
        self.data_items[data_item_lbl]['x_off'] = x_off
        self.data_items[data_item_lbl]['y_off'] = y_off
        self.data_items[data_item_lbl]['scale'] = scale

        self.update_marker(marker=self.marker_1)
        self.update_marker(marker=self.marker_2)

    def reset_scaling(self):
        data_item_lbl = self.main_added.currentText()
        org = self.data_items[data_item_lbl]['org_data']
        self.data_items[data_item_lbl]['data_item'].setData(deepcopy(org[0]), deepcopy(org[1]))
        self.change_without_emitting_signal(self.ec_offset_x, 0)
        self.change_without_emitting_signal(self.ec_offset_y, 0)
        self.change_without_emitting_signal(self.ec_scale, 1)
        self.change_without_emitting_signal(self.ec_normalize, False)
        self.data_items[data_item_lbl]['x_off'] = 0
        self.data_items[data_item_lbl]['y_off'] = 0
        self.data_items[data_item_lbl]['scale'] = 1
        self.data_items[data_item_lbl]['normalize'] = False

    def set_bgr_color(self, loading=False):
        if loading:
            cd_str = 'rgb' + str(self.plot_panel_design['bgr_color'].getRgb()[:3])
        else:
            cd = QColorDialog.getColor()
            cd_str = 'rgb' + str(cd.getRgb()[:3])
            self.plot_panel_design['bgr_color'] = cd
        self.ep_bgr_color.setStyleSheet(f'background-color: {cd_str}')

        self.set_plot_layout(loading=loading)

    def set_axes_color(self, loading=False):
        if loading:
            cd_str = 'rgb' + str(self.plot_panel_design['axes_color'].getRgb()[:3])
        else:
            cd = QColorDialog.getColor()
            cd_str = 'rgb' + str(cd.getRgb()[:3])
            self.plot_panel_design['axes_color'] = cd
        self.ep_axes_color.setStyleSheet(f'background-color: {cd_str}')

        self.set_plot_layout(loading=loading)

    def set_ticks_fsize(self, loading=False):
        self.set_plot_layout(loading=loading)

    def set_plot_layout(self, loading=False):
        bgr_clr = self.plot_panel_design['bgr_color']
        ax_clr = self.plot_panel_design['axes_color']

        if loading:
            self.ep_ticks_size.blockSignals(True)
            self.ep_labels_font_size.blockSignals(True)
            self.ep_ticks_size.setValue(self.plot_panel_design['ticks_fsize'].pointSize())
            self.ep_labels_font_size.setValue(self.plot_panel_design['labels_fsize'])
            self.ep_ticks_size.blockSignals(False)
            self.ep_labels_font_size.blockSignals(False)
        tfs = self.ep_ticks_size.value()
        self.plot_panel_design['ticks_fsize'].setPointSize(tfs)
        ticks_font = self.plot_panel_design['ticks_fsize']

        lbl_fsize = str(self.ep_labels_font_size.value()) + 'pt'
        lbl_clr = 'rgb' + str(ax_clr.getRgb()[:3])
        label_style = {'color': lbl_clr, 'font-size': lbl_fsize}
        self.plot_panel_design['labels_fsize'] = self.ep_labels_font_size.value()

        self.plot_panel.setBackground(bgr_clr)

        bottom_ax = self.plot_panel.plotItem.getAxis('bottom')
        bottom_ax.setPen(color=ax_clr)
        bottom_ax.setTextPen(color=ax_clr)
        bottom_ax.setTickFont(ticks_font)
        if not self.ep_xlabel.text() == '':
            xlbl = self.ep_xlabel.text().split(';')
            if len(xlbl) == 1:
                bottom_ax.setLabel(text=xlbl, **label_style)
            elif len(xlbl) == 2:
                bottom_ax.setLabel(text=xlbl[0], units=xlbl[1], **label_style)

        left_ax = self.plot_panel.plotItem.getAxis('left')
        left_ax.setPen(color=ax_clr)
        left_ax.setTextPen(color=ax_clr)
        left_ax.setTickFont(ticks_font)
        if not self.ep_ylabel.text() == '':
            ylbl = self.ep_ylabel.text().split(';')
            if len(ylbl) == 1:
                left_ax.setLabel(text=ylbl, **label_style)
            elif len(ylbl) == 2:
                left_ax.setLabel(text=ylbl[0], units=ylbl[1], **label_style)

    def get_data_limits(self):
        try:
            di_lbl = self.main_added.currentText()
            x, y = self.data_items[di_lbl]['data_item'].getData()
            xmin, xmax, ymin, ymax = x.min(), x.max(), y.min(), y.max()
        except KeyError:
            return

        for di in self.data_items.keys():
            x, y = self.data_items[di]['data_item'].getData()
            if x.min() < xmin:
                xmin = x.min()
            if x.max() > xmax:
                xmax = x.max()
            if y.min() < ymin:
                ymin = y.min()
            if y.max() > ymax:
                ymax = y.max()
        return xmin, xmax, ymin, ymax

    def drop_marker(self, marker=None):
        if len(self.data_items.keys()) == 0:
            no_data_item_box = QMessageBox()
            no_data_item_box.setIcon(QMessageBox.Information)
            no_data_item_box.setWindowTitle('Doh.')
            no_data_item_box.setText('No dataset to dump marker on.')
            no_data_item_box.setStandardButtons(QMessageBox.Ok)
            if no_data_item_box.exec() == QMessageBox.Ok:
                return

        marker = marker
        marker['dropped'] = not marker['dropped']
        if marker['dropped']:
            di = self.main_added.currentText()
            marker['dumped_at'] = di
            x, y = self.data_items[di]['data_item'].getData()
            marker['button'].setText('remove')
            marker['x'].setRange(x.min(), x.max())
            x0 = np.abs(x.min() - x.max()) / 2 + x.min()
            marker['x'].setValue(x0)
            marker['x'].setSingleStep(wp.get_step(x))
            x0_idx = wp.indexof(x0, x)
            y0 = y[x0_idx]
            marker['y'].setText('{:.2f}'.format(y0))
            self.set_markers_differences()
            xx, yy = [x0, x0], [y0, y0]
            if marker['idx'] == 1:
                s = 'star'
                mkp = pg.mkPen(color='r')
                b = pg.mkBrush(color='r')
            else:
                s = 'x'
                mkp = pg.mkPen(color='k')
                b = pg.mkBrush(color='k')
            marker['marker'].setData(xx, yy, symbol=s, pen=mkp, brush=b)
            self.plot_panel.addItem(marker['marker'])
        else:
            marker['button'].setText('drop')
            marker['x'].setValue(0)
            marker['y'].setText('-')
            self.plot_panel.removeItem(marker['marker'])
            self.set_markers_differences()

    def set_markers_differences(self):
        if self.marker_1['dropped'] and self.marker_2['dropped']:
            dx = np.abs(self.marker_1['x'].value() - self.marker_2['x'].value())
            y_m1 = float(self.marker_1['y'].text())
            y_m2 = float(self.marker_2['y'].text())
            dy = np.abs(y_m1 - y_m2)
            self.markers_dx.setText('{:.6f}'.format(dx))
            self.markers_dy.setText('{:.2f}'.format(dy))
        else:
            self.markers_dx.setText('-')
            self.markers_dy.setText('-')

    def update_marker(self, marker=None):
        marker = marker
        if marker['dropped']:
            x, y = self.data_items[marker['dumped_at']]['data_item'].getData()
            x0 = marker['x'].value()
            x_idx = wp.indexof(x0, x)
            y0 = y[x_idx]
            marker['y'].setText('{:.2f}'.format(y0))
            marker['marker'].setData([x0, x0], [y0, y0])
            self.set_markers_differences()
        else:
            return

    def set_annotation_color(self):
        name = self.ann_name.text()
        if not (name in self.annotations.keys()):
            no_annotation_box = QMessageBox()
            no_annotation_box.setIcon(QMessageBox.Information)
            no_annotation_box.setWindowTitle('Doh.')
            no_annotation_box.setText('No annotation to edit.')
            no_annotation_box.setStandardButtons(QMessageBox.Ok)
            if no_annotation_box.exec() == QMessageBox.Ok:
                return

        cd = QColorDialog.getColor()
        cd_str = cd.getRgb()[:3]
        self.ann_color.setStyleSheet(f'background-color: rgb{str(cd_str)}')
        self.annotations[name]['color'] = cd_str
        self.add_update_annotation()

    def add_update_annotation(self):
        name = self.ann_name.text()
        if name == '':
            no_name_box = QMessageBox()
            no_name_box.setIcon(QMessageBox.Information)
            no_name_box.setWindowTitle('Doh.')
            no_name_box.setText('Must specify name.')
            no_name_box.setStandardButtons(QMessageBox.Ok)
            if no_name_box.exec() == QMessageBox.Ok:
                return

        try:
            self.plot_panel.removeItem(self.annotations[name]['text_item'])
        except AttributeError:
            pass
        except KeyError:
            pass

        text = self.ann_text.text()
        x = self.ann_x.value()
        y = self.ann_y.value()
        fsize = self.ann_fsize.value()
        text_font = QFont()
        text_font.setPointSize(fsize)

        try:
            clr = self.annotations[name]['color']
        except KeyError:
            self.annotations[name] = {}
            self.ann_added.blockSignals(True)
            self.ann_added.addItem(name)
            self.ann_added.blockSignals(False)
            clr = 'k'
            self.annotations[name]['color'] = clr

        self.annotations[name]['text_item'] = pg.TextItem(text=text, anchor=(0, 1), color=clr)
        self.annotations[name]['text_item'].setFont(text_font)
        self.annotations[name]['text_item'].setPos(x, y)
        self.plot_panel.addItem(self.annotations[name]['text_item'])
        self.annotations[name]['text'] = text
        self.annotations[name]['fsize'] = fsize
        self.annotations[name]['x'] = x
        self.annotations[name]['y'] = y

    def remove_annotation(self):
        name = self.ann_name.text()
        try:
            self.plot_panel.removeItem(self.annotations[name]['text_item'])
            self.ann_added.removeItem(self.ann_added.currentIndex())
            del(self.annotations[name])
        except AttributeError:
            pass
        except KeyError:
            pass

    def ann_selection_changed(self):
        name = self.ann_added.currentText()
        if name == '':
            self.ann_name.setText('')
            self.ann_text.setText('')
            self.ann_fsize.setValue(14)
            self.ann_x.setValue(1)
            self.ann_y.setValue(1)
            self.ann_color.setStyleSheet('background-color: black')
        else:
            self.ann_name.setText(name)
            self.ann_text.setText(self.annotations[name]['text'])
            self.ann_fsize.setValue(self.annotations[name]['fsize'])
            self.ann_x.setValue(self.annotations[name]['x'])
            self.ann_y.setValue(self.annotations[name]['y'])
            clr = self.annotations[name]['color']
            self.ann_color.setStyleSheet(f'background-color: rgb{str(clr)}')

    @staticmethod
    def change_without_emitting_signal(widget, value):
        widget.blockSignals(True)
        try:
            if isinstance(widget, QCheckBox):
                widget.setChecked(value)
            elif isinstance(widget, QComboBox):
                widget.setCurrentText(value)
            else:
                widget.setValue(value)
        except AttributeError:
            print(f'Some error with setting value {value} in {widget}.')
        finally:
            widget.blockSignals(False)

    @staticmethod
    def set_qspinbox(box, box_range=[-1., 1.], value=1., decimals=3, step=0.1, max_w=80):
        box.setRange(box_range[0], box_range[1])
        box.setValue(value)
        box.setMaximumWidth(max_w)
        if isinstance(box, QDoubleSpinBox):
            box.setDecimals(decimals)
            box.setSingleStep(step)

    def save(self):
        save_selector_box = QMessageBox()
        save_selector_box.setIcon(QMessageBox.Question)
        save_selector_box.setWindowTitle('Save')
        save_selector_box.setText('Save an image or session?')
        save_selector_box.setStandardButtons(QMessageBox.Cancel | QMessageBox.No | QMessageBox.Yes)
        no_button = save_selector_box.button(QMessageBox.No)
        no_button.setText('Session')
        yes_button = save_selector_box.button(QMessageBox.Yes)
        yes_button.setText('Image')
        choice = save_selector_box.exec_()
        if choice == QMessageBox.Cancel:
            return
        elif choice == QMessageBox.No:
            self.save_session()
        elif choice == QMessageBox.Yes:
            self.save_image()

    def save_session(self):
        data_items_to_save = self.get_data_items_to_save()
        plot_design_to_save = self.get_plot_design_to_save()
        annotations_to_save = self.get_annotations_to_save()
        res = argparse.Namespace(
            data_items=data_items_to_save,
            plot_design=plot_design_to_save,
            annotations=annotations_to_save
        )

        full_path, types = QFileDialog.getSaveFileName(self, 'Save Session')

        dl.dump(res, full_path, force=True)

    def get_data_items_to_save(self):
        data_items_to_save = {}
        for key in self.data_items.keys():
            data_items_to_save[key] = {}
            for keyy in self.data_items[key].keys():
                if keyy == 'data_item':
                    data_items_to_save[key][keyy] = self.data_items[key][keyy].getData()
                else:
                    data_items_to_save[key][keyy] = self.data_items[key][keyy]
        return data_items_to_save

    def set_data_items_from_save(self, saved):
        self.data_items = {}
        for key in saved.keys():
            self.main_added.addItem(key)
            self.data_items[key] = {}
            for keyy in saved[key].keys():
                if keyy == 'data_item':
                    x, y = saved[key][keyy]
                    self.data_items[key][keyy] = CustomDataItem(x, y)
                else:
                    self.data_items[key][keyy] = saved[key][keyy]
            self.plot_panel.addItem(self.data_items[key]['data_item'])

    def get_plot_design_to_save(self):
        plot_design_to_save = {}
        for key in self.plot_panel_design.keys():
            if key == 'ticks_fsize':
                plot_design_to_save[key] = self.plot_panel_design[key].pointSize()
            else:
                plot_design_to_save[key] = self.plot_panel_design[key]
        return plot_design_to_save

    def set_plot_design_from_save(self, saved):
        self.plot_panel_design = {}
        for key in saved.keys():
            if key == 'ticks_fsize':
                self.plot_panel_design[key] = QFont()
                self.plot_panel_design[key].setPointSize(saved[key])
            else:
                self.plot_panel_design[key] = saved[key]

    def get_annotations_to_save(self):
        annotations_to_save = {}
        for key in self.annotations.keys():
            annotations_to_save[key] = {}
            for keyy in self.annotations[key].keys():
                if keyy == 'text_item':
                    pass
                else:
                    annotations_to_save[key][keyy] = self.annotations[key][keyy]
        return annotations_to_save

    def set_annotations_from_save(self, saved):
        self.annotations = {}
        for key in saved.keys():
            self.ann_added.addItem(key)
            self.annotations[key] = {}
            for keyy in saved[key].keys():
                if keyy == 'text':
                    self.annotations[key][keyy] = saved[key][keyy]
                else:
                    self.annotations[key][keyy] = saved[key][keyy]
                clr = saved[key]['color']
                self.annotations[key]['text_item'] = pg.TextItem(text=saved[key]['text'], anchor=(1, 0), color=clr)
        return self.annotations

    def load(self):
        warning_box = QMessageBox()
        warning_box.setIcon(QMessageBox.Information)
        warning_box.setWindowTitle('Load')
        warning_box.setText('Current progress will be lost.  Sure to continue?')
        warning_box.setStandardButtons(QMessageBox.Cancel | QMessageBox.Ok)
        choice = warning_box.exec_()
        if choice == QMessageBox.Ok:
            pass
        else:
            return

        full_path, types = QFileDialog.getOpenFileName(self, 'Load Session')
        session = dl.load_pickle(full_path)

        self.main_added.blockSignals(True)
        self.ann_added.blockSignals(True)
        del self.data_items
        del self.plot_panel_design
        del self.annotations
        self.main_added.clear()
        self.ann_added.clear()

        self.set_data_items_from_save(session.data_items)
        self.set_plot_design_from_save(session.plot_design)
        self.set_annotations_from_save(session.annotations)

        self.main_added.blockSignals(False)
        self.ann_added.blockSignals(False)

        for idx in range(self.main_added.count()):
            self.main_added.setCurrentIndex(idx)
            self.update_dataset(loading=True)
            self.set_pen(loading=True)

        self.set_bgr_color(loading=True)
        self.set_axes_color(loading=True)

        for idx in range(self.ann_added.count()):
            self.ann_added.setCurrentIndex(idx)
            self.add_update_annotation()

    def save_image(self):
        options = QFileDialog.Options()
        options |= QFileDialog.DontUseNativeDialog
        full_path, types = QFileDialog.getSaveFileName(
            self, 'Save Image', filter='Images (*.png *.jpg *.jpeg *.bmp)', options=options)
        to_save = pg.exporters.ImageExporter(self.plot_panel.plotItem)

        to_save.export(full_path)

    def closeEvent(self, event) :
        """ closeEvent is triggered on `Alt-F4` or mouse-click closing a 
        window. 
        """
        self.data_browser.thread[self.title].quit()
        self.data_browser.thread[self.title].wait()
        del(self.data_browser.thread[self.title])
        del(self.data_browser.plotting_tools[self.title])


class CustomDataItem(pg.PlotDataItem):

    def __init__(self, *args, **kwargs):
        super(CustomDataItem, self).__init__(*args, **kwargs)

        self.created = datetime.datetime.now()

